import pytest

from async_amqp.protocol import OPEN, CLOSED

from . import testcase


class TestClose(testcase.RabbitTestCase):
    @pytest.mark.trio
    @pytest.mark.xfail(reason="this triggers a cancellation error")
    async def test_close(self, amqp):
        assert amqp.state == OPEN
        # grab a ref here because py36 sets _stream_reader to None in
        # StreamReaderProtocol.connection_lost()
        sock = amqp._stream.socket
        await amqp.close()
        assert amqp.state == CLOSED
        assert sock.fileno() == -1
        # make sure those 2 tasks/futures are properly set as finished
        assert amqp.connection_closed.is_set()
        assert amqp._heartbeat_timer_recv is None
        assert amqp._heartbeat_timer_send is None

    @pytest.mark.trio
    @pytest.mark.xfail(reason="this triggers a cancellation error")
    async def test_multiple_close(self, amqp):
        # close is supposed to be idempotent
        await amqp.close()
        assert amqp.state == CLOSED
        await amqp.close()
        assert amqp.state == CLOSED



"""Connection tests"""

import socket
import pytest

from async_amqp.protocol import OPEN, CLOSED

from . import testcase
from anyio.abc import SocketAttribute


class TestAmqpConnection(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_connect(self, amqp):
        assert amqp.state == OPEN
        assert amqp.server_properties is not None

    @pytest.mark.trio
    async def test_connect_tuning(self):
        # frame_max should be higher than 131072
        self.reset_vhost()
        frame_max = 131072
        channel_max = 10
        heartbeat = 100
        proto = testcase.connect(
            host=self.host,
            port=self.port,
            virtualhost=self.vhost,
            channel_max=channel_max,
            frame_max=frame_max,
            heartbeat=heartbeat,
        )
        async with proto as amqp:
            assert amqp.state == OPEN, amqp.state
            assert amqp.server_properties is not None

            assert amqp.connection_tunning == {
                'frame_max': frame_max,
                'channel_max': channel_max,
                'heartbeat': heartbeat
            }

            assert amqp.server_channel_max == channel_max
            assert amqp.server_frame_max == frame_max
            assert amqp.server_heartbeat == heartbeat
        assert amqp.state == CLOSED, amqp.state

    @pytest.mark.trio
    async def test_socket_nodelay(self):
        self.reset_vhost()
        proto = testcase.connect(host=self.host, port=self.port, virtualhost=self.vhost)
        async with proto as amqp:
            sock = amqp._stream
            opt_val = sock.extra(SocketAttribute.raw_socket).getsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY)
            assert opt_val > 0



"""
    Amqp queue class tests
"""

import anyio
import pytest

from . import testcase
from async_amqp import exceptions


class TestQueueDeclare(testcase.RabbitTestCase):
    def setUp(self):
        super().setUp()
        self.consume_future = anyio.Event()

    async def callback(self, body, envelope, properties):
        await self.consume_future.set()
        self.consume_result = (body, envelope, properties)

    async def get_callback_result(self):
        await self.consume_future.wait()
        result = self.consume_result
        self.consume_future = anyio.Event()
        return result

    @pytest.mark.trio
    async def test_queue_declare_no_name(self, channel):
        result = await channel.queue_declare()
        assert result['queue'] is not None

    @pytest.mark.trio
    async def test_queue_declare(self, channel):
        queue_name = 'q15'
        result = await channel.queue_declare(queue_name)
        assert result['message_count'] == 0, result
        assert result['consumer_count'] == 0, result
        assert channel.protocol.local_name(result['queue']) == queue_name, result
        assert result

    @pytest.mark.trio
    async def test_queue_declare_passive(self, channel):
        queue_name = 'q16'
        await channel.queue_declare(queue_name)
        result = await channel.queue_declare(queue_name, passive=True)
        assert result['message_count'] == 0, result
        assert result['consumer_count'] == 0, result
        assert channel.protocol.local_name(result['queue']) == queue_name, result

    @pytest.mark.trio
    async def test_queue_declare_custom_x_message_ttl_32_bits(self, channel):
        queue_name = 'queue_name'
        # 2147483648 == 10000000000000000000000000000000
        # in binary, meaning it is 32 bit long
        x_message_ttl = 2147483648
        result = await channel.queue_declare(queue_name, arguments={
            'x-message-ttl': x_message_ttl
        })
        assert result['message_count'] == 0
        assert result['consumer_count'] == 0
        assert result['queue'].endswith(queue_name)
        assert result

    @pytest.mark.trio
    async def test_queue_declare_passive_nonexistant_queue(self, channel):
        queue_name = 'q17'
        with pytest.raises(exceptions.ChannelClosed) as cm:
            await channel.queue_declare(queue_name, passive=True)

        assert cm.value.code == 404

    @pytest.mark.trio
    async def test_wrong_parameter_queue(self, channel):
        queue_name = 'q18'
        await channel.queue_declare(queue_name, exclusive=False, auto_delete=False)

        with pytest.raises(exceptions.ChannelClosed) as cm:
            await channel.queue_declare(
                queue_name, passive=False, exclusive=True, auto_delete=True
            )

        assert cm.value.code in (405,406)

    @pytest.mark.trio
    async def test_multiple_channel_same_queue(self, amqp):
        queue_name = 'q19'

        async with amqp.new_channel() as channel1:
            async with amqp.new_channel() as channel2:

                result = await channel1.queue_declare(queue_name, passive=False)
                assert result['message_count'] == 0, result
                assert result['consumer_count'] == 0, result
                assert amqp.local_name(result['queue']) == queue_name, result

                result = await channel2.queue_declare(queue_name, passive=False)
                assert result['message_count'] == 0, result
                assert result['consumer_count'] == 0, result
                assert amqp.local_name(result['queue']) == queue_name, result

    async def _test_queue_declare(
        self, amqp, queue_name, exclusive=False, durable=False, auto_delete=False
    ):
        # declare queue
        async with amqp.new_channel() as channel:
            result = await channel.queue_declare(
                queue_name,
                no_wait=False,
                exclusive=exclusive,
                durable=durable,
                auto_delete=auto_delete
            )

            # assert that the returned results matches
            assert amqp.full_name(queue_name) == result['queue']

            queues = await self.list_queues(channel.protocol)
            queue = queues[queue_name]

            # assert queue has been declared with correct arguments
            assert queue_name == queue['name']
            assert 0 == queue.get('consumers', 0), queue
            assert 0 == queue.get('messages_ready', 0), queue
            assert auto_delete == queue['auto_delete'], queue
            assert durable == queue['durable'], queue

            # delete queue
            await self.safe_queue_delete(queue_name, channel)

    @pytest.mark.trio
    async def test_durable_and_auto_deleted(self, amqp):
        await self._test_queue_declare(amqp, 'q1', exclusive=False, durable=True, auto_delete=True)

    @pytest.mark.trio
    async def test_durable_and_not_auto_deleted(self, amqp):
        await self._test_queue_declare(
            amqp, 'q2', exclusive=False, durable=True, auto_delete=False
        )

    @pytest.mark.trio
    async def test_not_durable_and_auto_deleted(self, amqp):
        await self._test_queue_declare(
            amqp, 'q3', exclusive=False, durable=False, auto_delete=True
        )

    @pytest.mark.trio
    async def test_not_durable_and_not_auto_deleted(self, amqp):
        await self._test_queue_declare(
            amqp, 'q4', exclusive=False, durable=False, auto_delete=False
        )

    @pytest.mark.trio
    async def test_exclusive(self, amqp):
        async with amqp.new_channel() as channel:
            # create an exclusive queue
            await channel.queue_declare("q5", exclusive=True)
            # consume from it
            await channel.basic_consume(self.callback, queue_name="q5", no_wait=False)

            # create another amqp connection
            async with self.create_amqp() as amqp:
                async with amqp.new_channel() as channel:
                    # assert that this connection cannot connect to the queue
                    with pytest.raises(exceptions.ChannelClosed):
                        await channel.basic_consume(self.callback, queue_name="q5", no_wait=False)
                        # channels are auto deleted by test case

    @pytest.mark.trio
    async def test_not_exclusive(self, amqp):
        async with amqp.new_channel() as channel:
            # create a non-exclusive queue
            await channel.queue_declare('q6', exclusive=False)
            # consume it
            await channel.basic_consume(self.callback, queue_name='q6', no_wait=False)
            # create an other amqp connection
            async with self.create_amqp(test_seq=amqp.test_seq) as amqp2:
                async with amqp2.new_channel() as channel:
                    # assert that this connection can connect to the queue
                    await channel.basic_consume(self.callback, queue_name='q6', no_wait=False)


class TestQueueDelete(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_delete_queue(self, channel):
        queue_name = 'q7'
        await channel.queue_declare(queue_name)
        result = await channel.queue_delete(queue_name)
        assert result

    @pytest.mark.trio
    async def test_delete_inexistant_queue(self, amqp):
        queue_name = 'q8'
        async with amqp.new_channel() as channel:
            if self.server_version(amqp) < (3, 3, 5):
                with pytest.raises(exceptions.ChannelClosed) as cm:
                    result = await channel.queue_delete(queue_name)

                assert cm.value.code == 404

            else:
                result = await channel.queue_delete(queue_name)
                assert result


class TestQueueBind(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_bind_queue(self, channel):
        queue_name = 'q9'
        exchange_name = 'exchange_name'

        await channel.queue_declare(queue_name)
        await channel.exchange_declare(exchange_name, type_name='direct')

        result = await channel.queue_bind(queue_name, exchange_name, routing_key='')
        assert result

    @pytest.mark.trio
    async def test_bind_unexistant_exchange(self, channel):
        queue_name = 'q10'
        exchange_name = 'exchange_name'

        await channel.queue_declare(queue_name)
        with pytest.raises(exceptions.ChannelClosed) as cm:
            await channel.queue_bind(queue_name, exchange_name, routing_key='')
        assert cm.value.code == 404

    @pytest.mark.trio
    async def test_bind_unexistant_queue(self, channel):
        queue_name = 'q11'
        exchange_name = 'exchange_name'

        await channel.exchange_declare(exchange_name, type_name='direct')

        with pytest.raises(exceptions.ChannelClosed) as cm:
            await channel.queue_bind(queue_name, exchange_name, routing_key='')
        assert cm.value.code == 404

    @pytest.mark.trio
    async def test_unbind_queue(self, channel):
        queue_name = 'q12'
        exchange_name = 'exchange_name'

        await channel.queue_declare(queue_name)
        await channel.exchange_declare(exchange_name, type_name='direct')

        await channel.queue_bind(queue_name, exchange_name, routing_key='')

        result = await channel.queue_unbind(queue_name, exchange_name, routing_key='')
        assert result


class TestQueuePurge(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_purge_queue(self, channel):
        queue_name = 'q13'

        await channel.queue_declare(queue_name)
        result = await channel.queue_purge(queue_name)
        assert result['message_count'] == 0

    @pytest.mark.trio
    async def test_purge_queue_inexistant_queue(self, channel):
        queue_name = 'q14'

        with pytest.raises(exceptions.ChannelClosed) as cm:
            await channel.queue_purge(queue_name)
        assert cm.value.code == 404



"""
    Helper class to decode AMQP responses

AMQP Frame implementations


0      1         3         7                      size+7      size+8
+------+---------+---------+    +-------------+     +-----------+
| type | channel |   size  |    |   payload   |     | frame-end |
+------+---------+---------+    +-------------+     +-----------+
 octets   short     long         'size' octets          octet

The frame-end octet MUST always be the hexadecimal value %xCE

type:

Type = 1, "METHOD": method frame.
Type = 2, "HEADER": content header frame.
Type = 3, "BODY": content body frame.
Type = 4, "HEARTBEAT": heartbeat frame.


Method Payload

0          2           4
+----------+-----------+-------------- - -
| class-id | method-id | arguments...
+----------+-----------+-------------- - -
    short     short       ...

Content Payload

0          2        4           12               14
+----------+--------+-----------+----------------+------------- - -
| class-id | weight | body size | property flags | property list...
+----------+--------+-----------+----------------+------------- - -
   short     short    long long       short        remainder...

"""

import io
import struct
import socket
import os
import datetime
from anyio import ClosedResourceError, IncompleteRead
from itertools import count
from decimal import Decimal

import pamqp.encode
import pamqp.frame

from . import exceptions
from . import constants as amqp_constants
from .properties import Properties

DUMP_FRAMES = False


async def read(reader):
    """Read a new frame from the wire

        reader:     anyio Stream

    Returns (channel, frame) a tuple containing both channel and the pamqp frame,
                             the object describing the frame
    """
    if not reader:
        raise exceptions.AmqpClosedConnection()
    try:
        data = await reader.receive_exactly(7)
    except (ClosedResourceError, IncompleteRead) as ex:
        raise exceptions.AmqpClosedConnection() from ex

    frame_type, channel, frame_length = pamqp.frame.frame_parts(data)

    payload_data = await reader.receive_exactly(frame_length)
    frame = None

    if frame_type == amqp_constants.TYPE_METHOD:
        frame = pamqp.frame._unmarshal_method_frame(payload_data)

    elif frame_type == amqp_constants.TYPE_HEADER:
        frame = pamqp.frame._unmarshal_header_frame(payload_data)

    elif frame_type == amqp_constants.TYPE_BODY:
        frame = pamqp.frame._unmarshal_body_frame(payload_data)

    elif frame_type == amqp_constants.TYPE_HEARTBEAT:
        frame = pamqp.heartbeat.Heartbeat()

    frame_end = await reader.receive_exactly(1)
    assert frame_end == amqp_constants.FRAME_END
    return channel, frame



"""
    Delivery of messages to consumers
"""


class Envelope:
    """Class for basic deliver message fields"""
    __slots__ = ('consumer_tag', 'delivery_tag', 'exchange_name', 'routing_key', 'is_redeliver')

    def __init__(self, consumer_tag, delivery_tag, exchange_name, routing_key, is_redeliver):
        self.consumer_tag = consumer_tag
        self.delivery_tag = delivery_tag
        self.exchange_name = exchange_name
        self.routing_key = routing_key
        self.is_redeliver = is_redeliver


class ReturnEnvelope:
    """ Class for basic return message fields"""
    __slots__ = ('reply_code', 'reply_text', 'exchange_name', 'routing_key')

    def __init__(self, reply_code, reply_text, exchange_name, routing_key):
        self.reply_code = reply_code
        self.reply_text = reply_text
        self.exchange_name = exchange_name
        self.routing_key = routing_key



"""
    Tests for message properties for basic deliver
"""

import anyio
import pytest
import logging

from . import testcase

logger = logging.getLogger(__name__)

server_queue_name = 'server_queue'
client_queue_name = 'client_reply_queue'
exchange_name = 'exchange_name'
server_routing_key = 'reply_test'


class TestReplyOld(testcase.RabbitTestCase):
    """RPC test using classic callbacks"""

    async def _server(
        self,
        amqp,
        server_future,
        exchange_name,
        routing_key,
        *,
        task_status,
    ):
        """Consume messages and reply to them by publishing messages back
        to the client using routing key set to the reply_to property
        """
        async with amqp.new_channel() as channel:
            await channel.queue_declare(server_queue_name, exclusive=False, no_wait=False)
            await channel.exchange_declare(exchange_name, type_name='direct')
            await channel.queue_bind(server_queue_name, exchange_name, routing_key=routing_key)

            async def server_callback(channel, body, envelope, properties):
                logger.debug('Server received message')
                publish_properties = {'correlation_id': properties.correlation_id}
                logger.debug('Replying to %r', properties.reply_to)
                await channel.publish(
                    b'reply message', exchange_name, properties.reply_to, publish_properties
                )
                server_future.test_result = (body, envelope, properties)
                server_future.set()
                logger.debug('Server replied')

            await channel.basic_consume(server_callback, queue_name=server_queue_name)
            logger.debug('Server consuming messages')
            task_status.started()
            await server_future.wait()

    async def _client(
        self,
        amqp,
        client_future,
        exchange_name,
        server_routing_key,
        correlation_id,
        client_routing_key,
        *,
        task_status,
    ):
        """Declare a queue, bind client_routing_key to it, and publish a
        message to the server with the reply_to property set to that
        routing key
        """
        async with amqp.new_channel() as client_channel:
            await client_channel.queue_declare(client_queue_name, exclusive=True, no_wait=False)
            await client_channel.queue_bind(
                client_queue_name, exchange_name, routing_key=client_routing_key
            )

            async def client_callback(channel, body, envelope, properties):
                logger.debug('Client received message')
                client_future.test_result = (body, envelope, properties)
                client_future.set()

            await client_channel.basic_consume(client_callback, queue_name=client_queue_name)
            logger.debug('Client consuming messages')
            task_status.started()

            await client_channel.publish(
                b'client message', exchange_name, server_routing_key, {
                    'correlation_id': correlation_id,
                    'reply_to': client_routing_key
                }
            )
            logger.debug('Client published message')
            await client_future.wait()

    @pytest.mark.trio
    async def test_reply_to(self, amqp):
        server_future = anyio.Event()
        async with anyio.create_task_group() as n:
            await n.start(self._server, amqp, server_future, exchange_name, server_routing_key)

            correlation_id = 'secret correlation id'
            client_routing_key = 'secret_client_key'

            client_future = anyio.Event()
            await n.start(
                self._client, amqp, client_future, exchange_name, server_routing_key,
                correlation_id, client_routing_key
            )

            logger.debug('Waiting for server to receive message')
            await server_future.wait()
            server_body, server_envelope, server_properties = \
                server_future.test_result
            assert server_body == b'client message'
            assert server_properties.correlation_id == correlation_id
            assert server_properties.reply_to == client_routing_key
            assert server_envelope.routing_key == server_routing_key

            logger.debug('Waiting for client to receive message')
            await client_future.wait()
            client_body, client_envelope, client_properties = \
                client_future.test_result
            assert client_body == b'reply message'
            assert client_properties.correlation_id == correlation_id
            assert client_envelope.routing_key == client_routing_key
            n.cancel_scope.cancel()


class TestReplyNew(testcase.RabbitTestCase):
    """RPC test using iteration"""

    async def _server(
        self,
        amqp,
        server_future,
        exchange_name,
        routing_key,
        *,
        task_status,
    ):
        """Consume messages and reply to them by publishing messages back
        to the client using routing key set to the reply_to property
        """
        async with amqp.new_channel() as channel:
            await channel.queue_declare(server_queue_name, exclusive=False, no_wait=False)
            await channel.exchange_declare(exchange_name, type_name='direct')
            await channel.queue_bind(server_queue_name, exchange_name, routing_key=routing_key)

            async with anyio.create_task_group() as n:
                await n.start(self._server_consumer, channel, server_future)
                task_status.started()
                await server_future.wait()
                self._server_scope.cancel()

    async def _server_consumer(self, channel, server_future, *, task_status):
        with anyio.CancelScope() as scope:
            self._server_scope = scope
            async with channel.new_consumer(queue_name=server_queue_name) \
                    as data:
                logger.debug('Server consuming messages')
                task_status.started()
                async for body, envelope, properties in data:

                    logger.debug('Server received message')
                    publish_properties = {'correlation_id': properties.correlation_id}
                    logger.debug('Replying to %r', properties.reply_to)
                    await channel.publish(
                        b'reply message', exchange_name, properties.reply_to, publish_properties
                    )
                    server_future.test_result = (body, envelope, properties)
                    server_future.set()
                    logger.debug('Server replied')

    async def _client(
        self,
        amqp,
        client_future,
        exchange_name,
        server_routing_key,
        correlation_id,
        client_routing_key,
        *,
        task_status,
    ):
        """Declare a queue, bind client_routing_key to it, and publish a
        message to the server with the reply_to property set to that
        routing key
        """
        async with amqp.new_channel() as client_channel:
            await client_channel.queue_declare(client_queue_name, exclusive=True, no_wait=False)
            await client_channel.queue_bind(
                client_queue_name, exchange_name, routing_key=client_routing_key
            )

            async with anyio.create_task_group() as n:
                await n.start(self._client_consumer, client_channel, client_future)
                task_status.started()

                await client_channel.publish(
                    b'client message', exchange_name, server_routing_key, {
                        'correlation_id': correlation_id,
                        'reply_to': client_routing_key
                    }
                )
                logger.debug('Client published message')
                await client_future.wait()
                self._client_scope.cancel()

    async def _client_consumer(self, channel, client_future, *, task_status):
        with anyio.CancelScope() as scope:
            self._client_scope = scope
            async with channel.new_consumer(queue_name=client_queue_name) \
                    as data:
                task_status.started()
                logger.debug('Client consuming messages')

                async for body, envelope, properties in data:
                    logger.debug('Client received message')
                    client_future.test_result = (body, envelope, properties)
                    client_future.set()

    @pytest.mark.trio
    async def test_reply_to(self, amqp):
        server_future = anyio.Event()
        async with anyio.create_task_group() as n:
            await n.start(self._server, amqp, server_future, exchange_name, server_routing_key)

            correlation_id = 'secret correlation id'
            client_routing_key = 'secret_client_key'

            client_future = anyio.Event()
            await n.start(
                self._client, amqp, client_future, exchange_name, server_routing_key,
                correlation_id, client_routing_key
            )

            logger.debug('Waiting for server to receive message')
            await server_future.wait()
            server_body, server_envelope, server_properties = \
                server_future.test_result
            assert server_body == b'client message'
            assert server_properties.correlation_id == correlation_id
            assert server_properties.reply_to == client_routing_key
            assert server_envelope.routing_key == server_routing_key

            logger.debug('Waiting for client to receive message')
            await client_future.wait()
            client_body, client_envelope, client_properties = \
                client_future.test_result
            assert client_body == b'reply message'
            assert client_properties.correlation_id == correlation_id
            assert client_envelope.routing_key == client_routing_key
            n.cancel_scope.cancel()



#!/usr/bin/env python
"""
    Rabbitmq.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-four-python.html

"""

import anyio
import async_amqp

import sys


async def callback(channel, body, envelope, properties):
    print("consumer {} recved {} ({})".format(envelope.consumer_tag, body, envelope.delivery_tag))

async def receive_log():
    try:
        async with async_amqp.connect_amqp() as protocol:

            channel = await protocol.channel()
            exchange_name = 'direct_logs'

            await channel.exchange(exchange_name, 'direct')

            result = await channel.queue(
                queue_name='', durable=False, auto_delete=True
            )

            queue_name = result['queue']

            severities = sys.argv[1:]
            if not severities:
                print("Usage: %s [info] [warning] [error]" % (sys.argv[0],))
                sys.exit(1)

            for severity in severities:
                await channel.queue_bind(
                    exchange_name='direct_logs',
                    queue_name=queue_name,
                    routing_key=severity,
                )

            print(' [*] Waiting for logs. To exit press CTRL+C')

            with anyio.fail_after(10):
                await channel.basic_consume(callback, queue_name=queue_name)

    except async_amqp.AmqpClosedConnection:
        print("closed connections")
        return


anyio.run(receive_log)



import reprlib
reprlib.aRepr.maxtuple = 999
reprlib.aRepr.maxlist = 999
reprlib.aRepr.maxarray = 999
reprlib.aRepr.maxdict = 999
reprlib.aRepr.maxset = 999
reprlib.aRepr.maxfrozenset = 999
reprlib.aRepr.maxdeque = 999
reprlib.aRepr.maxstring = 999
reprlib.aRepr.maxlong = 999
reprlib.aRepr.maxother = 999



#!/usr/bin/env python
"""
    RPC client, async_amqp implementation of RPC examples from RabbitMQ tutorial

"""

import anyio
import uuid

import async_amqp


class FibonacciRpcClient(object):
    def __init__(self):
        self.protocol = None
        self.channel = None
        self.callback_queue = None
        self.waiter = anyio.Event()

    async def connect(self, channel):
        """ an `__init__` method can't be a coroutine"""
        result = await channel.queue_declare(
            queue_name='', exclusive=True
        )
        self.callback_queue = result['queue']

        await channel.basic_consume(
            self.on_response,
            no_ack=True,
            queue_name=self.callback_queue,
        )

    async def on_response(self, channel, body, envelope, properties):
        if self.corr_id == properties.correlation_id:
            self.response = body

        await self.waiter.set()

    async def call(self, n):
        async with async_amqp.connect_amqp() as protocol:
            async with protocol.channel() as channel:
                await self.connect(channel)

                self.response = None
                self.corr_id = str(uuid.uuid4())
                await self.channel.basic_publish(
                    payload=str(n),
                    exchange_name='',
                    routing_key='rpc_queue',
                    properties={
                        'reply_to': self.callback_queue,
                        'correlation_id': self.corr_id,
                    },
                )
                await self.waiter.wait()

                return int(self.response)


async def rpc_client():
    fibonacci_rpc = FibonacciRpcClient()
    print(" [x] Requesting fib(30)")
    response = await fibonacci_rpc.call(30)
    print(" [.] Got %r" % response)


anyio.run(rpc_client)



"""
    Tests the "Channel" amqp class implementation
"""

import os
import pytest

from . import testcase
from async_amqp import exceptions

IMPLEMENT_CHANNEL_FLOW = os.environ.get('IMPLEMENT_CHANNEL_FLOW', False)


class TestChannel(testcase.RabbitTestCase):

    _multiprocess_can_split_ = True

    @pytest.mark.trio
    async def test_open(self, amqp):
        async with amqp.new_channel() as channel:
            assert channel.channel_id != 0
            assert channel.is_open

    @pytest.mark.trio
    async def test_close(self, amqp):
        async with amqp.new_channel() as channel:
            assert channel.is_open
        assert not channel.is_open

    @pytest.mark.trio
    async def test_server_initiated_close(self, channel):
        try:
            await channel.basic_get(queue_name='non-existant')
        except exceptions.ChannelClosed as e:
            assert e.code == 404
        assert not channel.is_open

    @pytest.mark.trio
    async def test_alreadyclosed_channel(self, channel):
        result = await channel.close()
        assert result is True

        with pytest.raises(exceptions.ChannelClosed):
            result = await channel.close()

    @pytest.mark.trio
    async def test_multiple_open(self, amqp):
        async with amqp.new_channel() as channel1:
            async with amqp.new_channel() as channel2:
                assert channel1.channel_id != channel2.channel_id

    @pytest.mark.trio
    async def test_channel_active_flow(self, channel):
        result = await channel.flow(active=True)
        assert result['active']

    @pytest.mark.skipif(
        IMPLEMENT_CHANNEL_FLOW is False, reason="active=false is not implemented in RabbitMQ"
    )
    @pytest.mark.trio
    async def test_channel_inactive_flow(self, channel):
        result = await channel.flow(active=False)
        assert not result['active']
        result = await channel.flow(active=True)

    @pytest.mark.trio
    async def test_channel_active_flow_twice(self, channel):
        result = await channel.flow(active=True)
        assert result['active']
        result = await channel.flow(active=True)

    @pytest.mark.skipif(
        IMPLEMENT_CHANNEL_FLOW is False, reason="active=false is not implemented in RabbitMQ"
    )
    @pytest.mark.trio
    async def test_channel_active_inactive_flow(self, channel):
        result = await channel.flow(active=True)
        assert result['active']
        result = await channel.flow(active=False)
        assert not result['active']


class TestChannelId(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_channel_id_release_close(self, amqp):
        channels_count_start = amqp.channels_ids_count
        async with amqp.new_channel() as channel:
            assert amqp.channels_ids_count == channels_count_start + 1
        assert not channel.is_open
        assert amqp.channels_ids_count == channels_count_start



"""
    Amqp channel specification
"""

import sys
import anyio
import logging
import uuid
import io
import inspect
from itertools import count

import pamqp

from . import constants as amqp_constants
from . import frame as amqp_frame
from . import exceptions
from . import properties as amqp_properties
from .envelope import Envelope, ReturnEnvelope
from .future import Future
from .exceptions import AmqpClosedConnection, SynchronizationError

logger = logging.getLogger(__name__)


class BasicListener:
    """This class is returned by :meth:Channel.new_consumer`.
    It is responsible for telling AMQP to start sending data.
    """

    def __init__(self, channel, consumer_tag, **kwargs):
        self.channel = channel
        self.kwargs = kwargs
        self.consumer_tag = consumer_tag

    async def _data(self, channel, msg, env, prop):
        if msg is None:
            await self._q_w.send(None)
        else:
            await self._q_w.send((msg, env, prop))

    if sys.version_info >= (3,5,3):
        def __aiter__(self):
            return self
    else:
        async def __aiter__(self):
            return self

    async def __anext__(self):
        res = await self._q_r.receive()
        if res is None:
            raise StopAsyncIteration
        return res

    async def get(self):
        return await self._q_r.receive()

    async def __aenter__(self):
        await self.channel.basic_consume(self._data, consumer_tag=self.consumer_tag, **self.kwargs)
        self._q_w,self._q_r = anyio.create_memory_object_stream(30)  # TODO: 2 + possible prefetch
        return self

    async def __aexit__(self, *tb):
        with anyio.CancelScope(shield=True):
            try:
                await self.channel.basic_cancel(self.consumer_tag)
            except AmqpClosedConnection:
                pass
        await self._q_w.aclose()
        await self._q_r.aclose()
        # these messages are not acknowledged, thus deleting the queue will
        # not lose them

    def __enter__(self):
        raise RuntimeError("You need to use 'async with'.")

    def __exit__(self, *tb):
        raise RuntimeError("You need to use 'async with'.")

    def __iter__(self):
        raise RuntimeError("You need to use 'async for'.")


class Channel:
    _q_w,_q_r = None,None # for returned messages

    def __init__(self, protocol, channel_id):
        self.protocol = protocol
        self.channel_id = channel_id
        self.consumer_queues = {}
        self.consumer_callbacks = {}
        self.response_future = None
        self.close_event = anyio.Event()
        self.cancelled_consumers = set()
        self.last_consumer_tag = None
        self.publisher_confirms = False

        self.delivery_tag_iter = None
        # counting iterator, used for mapping delivered messages
        # to publisher confirms

        self._write_lock = anyio.Lock()
        self._exchange_declare_lock = anyio.Lock()
        self._queue_bind_lock = anyio.Lock()

        self._futures = {}
        self._ctag_events = {}

    def __aiter__(self):
        if self._q_w is None:
            self._q_w,self._q_r = anyio.create_memory_object_stream(30)  # TODO: 2 + possible prefetch
        return self

    if sys.version_info < (3,5,3):
        _aiter = __aiter__
        async def __aiter__(self):
            return self._aiter()

    async def __anext__(self):
        res = await self._q_r.receive()
        if res is None:
            raise StopAsyncIteration
        return res

    def _add_future(self, fut):
        self._futures[fut.rpc_name] = fut

    def _set_waiter(self, rpc_name):
        if rpc_name in self._futures:
            raise exceptions.SynchronizationError("Waiter already exists")

        return Future(self, rpc_name)

    def _get_waiter(self, rpc_name):
        fut = self._futures.pop(rpc_name, None)
        if not fut:
            raise exceptions.SynchronizationError("Call %r didn't set a waiter" % rpc_name)
        return fut

    @property
    def is_open(self):
        if self.protocol.connection_closed.is_set():
            return False
        return not self.close_event.is_set()

    async def connection_closed(self, server_code=None, server_reason=None, exception=None):
        for future in self._futures.values():
            if future.done():
                continue
            if exception is None:
                kwargs = {}
                if server_code is not None:
                    kwargs['code'] = server_code
                if server_reason is not None:
                    kwargs['message'] = server_reason
                exception = exceptions.ChannelClosed(**kwargs)
            future.set_exception(exception)

        self.protocol.release_channel_id(self.channel_id)
        self.close_event.set()
        if self._q_w is not None:
            await self._q_w.aclose()

    async def dispatch_frame(self, frame):
        methods = {
            pamqp.commands.Channel.OpenOk.name: self.open_ok,
            pamqp.commands.Channel.FlowOk.name: self.flow_ok,
            pamqp.commands.Channel.CloseOk.name: self.close_ok,
            pamqp.commands.Channel.Close.name: self.server_channel_close,

            pamqp.commands.Exchange.DeclareOk.name: self.exchange_declare_ok,
            pamqp.commands.Exchange.BindOk.name: self.exchange_bind_ok,
            pamqp.commands.Exchange.UnbindOk.name: self.exchange_unbind_ok,
            pamqp.commands.Exchange.DeleteOk.name: self.exchange_delete_ok,

            pamqp.commands.Queue.DeclareOk.name: self.queue_declare_ok,
            pamqp.commands.Queue.DeleteOk.name: self.queue_delete_ok,
            pamqp.commands.Queue.BindOk.name: self.queue_bind_ok,
            pamqp.commands.Queue.UnbindOk.name: self.queue_unbind_ok,
            pamqp.commands.Queue.PurgeOk.name: self.queue_purge_ok,

            pamqp.commands.Basic.QosOk.name: self.basic_qos_ok,
            pamqp.commands.Basic.ConsumeOk.name: self.basic_consume_ok,
            pamqp.commands.Basic.CancelOk.name: self.basic_cancel_ok,
            pamqp.commands.Basic.GetOk.name: self.basic_get_ok,
            pamqp.commands.Basic.GetEmpty.name: self.basic_get_empty,
            pamqp.commands.Basic.Deliver.name: self.basic_deliver,
            pamqp.commands.Basic.Cancel.name: self.server_basic_cancel,
            pamqp.commands.Basic.Ack.name: self.basic_server_ack,
            pamqp.commands.Basic.Nack.name: self.basic_server_nack,
            pamqp.commands.Basic.RecoverOk.name: self.basic_recover_ok,
            pamqp.commands.Basic.Return.name: self.basic_return,

            pamqp.commands.Confirm.SelectOk.name: self.confirm_select_ok,
        }

        if frame.name not in methods:
            raise NotImplementedError("Frame %s is not implemented" % frame.name)

        await methods[frame.name](frame)

    async def _write_frame(self, frame, request, check_open=True, drain=True):
        await self.protocol.ensure_open()
        if not self.is_open and check_open:
            raise exceptions.ChannelClosed()
        await self.protocol._write_frame(frame, request)
        if drain:
            await self.protocol._drain()

    async def _write_frame_awaiting_response(
        self, waiter_id, channel_id, request, no_wait, check_open=True, drain=True
    ):
        '''Write a frame and set a waiter for the response
        (unless no_wait is set)
        '''
        if no_wait:
            await self._write_frame(channel_id, request, check_open=check_open, drain=drain)
            return None

        async with self._write_lock:
            f = self._set_waiter(waiter_id)
            try:
                await self._write_frame(channel_id, request, check_open=check_open, drain=drain)
            except BaseException as exc:
                self._get_waiter(waiter_id)
                f.cancel()
                raise
            return await f()

#
# Channel class implementation
#

    async def open(self):
        """Open the channel on the server."""
        request = pamqp.commands.Channel.Open()
        return await self._write_frame_awaiting_response('open', self.channel_id, request, no_wait=False, check_open=False)

    async def open_ok(self, frame):
        self.close_event = anyio.Event()
        fut = self._get_waiter('open')
        fut.set_result(True)
        logger.debug("Channel is open")

    async def close(self, reply_code=0, reply_text="Normal Shutdown"):
        """Close the channel."""
        if not self.is_open:
            raise exceptions.ChannelClosed("channel already closed or closing")
        self.close_event.set()
        if self._q_w is not None:
            await self._q_w.aclose()
        request = pamqp.commands.Channel.Close(reply_code, reply_text, class_id=0, method_id=0)
        return await self._write_frame_awaiting_response('close', self.channel_id, request, no_wait=False, check_open=False)

    async def close_ok(self, frame):
        try:
            w = self._get_waiter('close')
        except SynchronizationError:
            pass
        else:
            w.set_result(True)
        logger.info("Channel closed")
        self.protocol.release_channel_id(self.channel_id)

    async def _send_channel_close_ok(self):
        request = pamqp.commands.Channel.CloseOk()
        # intentionally not locked
        await self._write_frame(self.channel_id, request)

    async def server_channel_close(self, frame):
        try:
            await self._send_channel_close_ok()
        except exceptions.ChannelClosed:
            pass
        results = {
            'reply_code': frame.reply_code,
            'reply_text': frame.reply_text,
            'class_id': frame.class_id,
            'method_id': frame.method_id,
        }
        await self.connection_closed(results['reply_code'], results['reply_text'])

    async def flow(self, active):
        request = pamqp.commands.Channel.Flow(active)
        return await self._write_frame_awaiting_response('flow', self.channel_id, request, no_wait=False, check_open=False)

    async def flow_ok(self, frame):
        self.close_event = anyio.Event()
        fut = self._get_waiter('flow')
        fut.set_result({'active': frame.active})

        logger.debug("Flow ok")

#
# Exchange class implementation
#

    async def exchange_declare(
        self,
        exchange_name,
        type_name,
        passive=False,
        durable=False,
        auto_delete=False,
        no_wait=False,
        arguments=None
    ):
        request = pamqp.commands.Exchange.Declare(
            exchange=exchange_name,
            exchange_type=type_name,
            passive=passive,
            durable=durable,
            auto_delete=auto_delete,
            nowait=no_wait,
            arguments=arguments
        )

        async with self._exchange_declare_lock:
            return await self._write_frame_awaiting_response('exchange_declare', self.channel_id, request, no_wait)

    async def exchange_declare_ok(self, frame):
        future = self._get_waiter('exchange_declare')
        future.set_result(True)
        logger.debug("Exchange declared")
        return future

    async def exchange_delete(self, exchange_name, if_unused=False, no_wait=False):
        request = pamqp.commands.Exchange.Delete(exchange=exchange_name, if_unused=if_unused, nowait=no_wait)

        return await self._write_frame_awaiting_response('exchange_delete', self.channel_id, request, no_wait)

    async def exchange_delete_ok(self, frame):
        future = self._get_waiter('exchange_delete')
        future.set_result(True)
        logger.debug("Exchange deleted")

    async def exchange_bind(
        self, exchange_destination, exchange_source, routing_key, no_wait=False, arguments=None
    ):
        if arguments is None:
            arguments = {}
        request = pamqp.commands.Exchange.Bind(
            destination=exchange_destination,
            source=exchange_source,
            routing_key=routing_key,
            nowait=no_wait,
            arguments=arguments
        )
        return await self._write_frame_awaiting_response('exchange_bind', self.channel_id, request, no_wait)

    async def exchange_bind_ok(self, frame):
        future = self._get_waiter('exchange_bind')
        future.set_result(True)
        logger.debug("Exchange bound")

    async def exchange_unbind(
        self, exchange_destination, exchange_source, routing_key, no_wait=False, arguments=None
    ):
        if arguments is None:
            arguments = {}

        request = pamqp.commands.Exchange.Unbind(
            destination=exchange_destination,
            source=exchange_source,
            routing_key=routing_key,
            nowait=no_wait,
            arguments=arguments,
        )
        return await self._write_frame_awaiting_response('exchange_unbind', self.channel_id, request, no_wait)

    async def exchange_unbind_ok(self, frame):
        future = self._get_waiter('exchange_unbind')
        future.set_result(True)
        logger.debug("Exchange bound")

#
# Queue class implementation
#

    async def queue_declare(
        self,
        queue_name=None,
        passive=False,
        durable=False,
        exclusive=False,
        auto_delete=False,
        no_wait=False,
        arguments=None
    ):
        """Create or check a queue on the broker
           Args:
                queue_name:
                    str, the queue to receive message from. The server generate
                    a queue_name if not specified.
                passive:
                    bool, if set, the server will reply with Declare-Ok if
                    the queue already exists with the same name, and raise
                    an error if not. Checks for the same parameter as well.
                durable:
                    bool: If set when creating a new queue, the queue will
                    be marked as durable. Durable queues remain active when
                    a server restarts.
                exclusive:
                    bool, request exclusive consumer access, meaning only
                    this consumer can access the queue
                no_wait:
                    bool, if set, the server will not respond to the method
                arguments:
                    dict, AMQP arguments to be passed when creating the
                    queue.
        """
        if arguments is None:
            arguments = {}

        if not queue_name:
            queue_name = 'async_amqp.gen-'  + str(uuid.uuid4())
        request = pamqp.commands.Queue.Declare(
            queue=queue_name,
            passive=passive,
            durable=durable,
            exclusive=exclusive,
            auto_delete=auto_delete,
            nowait=no_wait,
            arguments=arguments
        )
        return await self._write_frame_awaiting_response(
            'queue_declare' + queue_name, self.channel_id, request, no_wait)

    async def queue_declare_ok(self, frame):
        results = {
            'queue': frame.queue,
            'message_count': frame.message_count,
            'consumer_count': frame.consumer_count,
        }
        future = self._get_waiter('queue_declare' + results['queue'])
        future.set_result(results)
        logger.debug("Queue declared")

    async def queue_delete(self, queue_name, if_unused=False, if_empty=False, no_wait=False):
        """Delete a queue in RabbitMQ
            Args:
                queue_name:
                    str, the queue to receive message from
                if_unused:
                    bool, the queue is deleted if it has no consumers.
                    Raise if not.
                if_empty:
                    bool, the queue is deleted if it has no messages. Raise
                    if not.
                no_wait:
                    bool, if set, the server will not respond to the method
        """
        request = pamqp.commands.Queue.Delete(
            queue=queue_name,
            if_unused=if_unused,
            if_empty=if_empty,
            nowait=no_wait
        )
        return await self._write_frame_awaiting_response('queue_delete', self.channel_id, request, no_wait)

    async def queue_delete_ok(self, frame):
        future = self._get_waiter('queue_delete')
        future.set_result(True)
        logger.debug("Queue deleted")

    async def queue_bind(
        self, queue_name, exchange_name, routing_key, no_wait=False, arguments=None
    ):
        """Bind a queue to an exchange."""
        if arguments is None:
            arguments = {}
        request = pamqp.commands.Queue.Bind(
            queue=queue_name,
            exchange=exchange_name,
            routing_key=routing_key,
            nowait=no_wait,
            arguments=arguments
        )
        async with self._queue_bind_lock:
            return await self._write_frame_awaiting_response('queue_bind', self.channel_id, request, no_wait)

    async def queue_bind_ok(self, frame):
        future = self._get_waiter('queue_bind')
        future.set_result(True)
        logger.debug("Queue bound")

    async def queue_unbind(self, queue_name, exchange_name, routing_key, arguments=None):
        if arguments is None:
            arguments = {}
        request = pamqp.commands.Queue.Unbind(
            queue=queue_name,
            exchange=exchange_name,
            routing_key=routing_key,
            arguments=arguments
        )
        async with self._queue_bind_lock:
            return await self._write_frame_awaiting_response('queue_unbind', self.channel_id, request, no_wait=False)

    async def queue_unbind_ok(self, frame):
        future = self._get_waiter('queue_unbind')
        future.set_result(True)
        logger.debug("Queue unbound")

    async def queue_purge(self, queue_name, no_wait=False):
        request = pamqp.commands.Queue.Purge(
            queue=queue_name, nowait=no_wait
        )
        return await self._write_frame_awaiting_response('queue_purge', self.channel_id, request, no_wait=no_wait)

    async def queue_purge_ok(self, frame):
        future = self._get_waiter('queue_purge')
        future.set_result({'message_count': frame.message_count})

#
# Basic class implementation
#

    async def basic_publish(
        self,
        payload,
        exchange_name,
        routing_key,
        properties=None,
        mandatory=False,
        immediate=False
    ):
        async with self._write_lock:
            if properties is None:
                properties = {}

            method_request = pamqp.commands.Basic.Publish(
                exchange=exchange_name,
                routing_key=routing_key,
                mandatory=mandatory,
                immediate=immediate
            )

            await self._write_frame(self.channel_id, method_request, drain=False)

            header_request = pamqp.header.ContentHeader(
                body_size=len(payload),
                properties=pamqp.commands.Basic.Properties(**properties)
            )
            await self._write_frame(self.channel_id, header_request, drain=False)

            frame_max = self.protocol.server_frame_max or len(payload)
            for chunk in (payload[0 + i:frame_max + i] for i in range(0, len(payload), frame_max)):

                content_request = pamqp.body.ContentBody(chunk)
                await self._write_frame(self.channel_id, content_request, drain=False)

        await self.protocol._drain()

    async def basic_qos(self, prefetch_size=0, prefetch_count=0, connection_global=False):
        """Specifies quality of service.

        Args:
            prefetch_size:
                int, request that messages be sent in advance so that when
                the client finishes processing a message, the following
                message is already held locally
            prefetch_count:
                int: Specifies a prefetch window in terms of whole
                messages. This field may be used in combination with the
                prefetch-size field; a message will only be sent in advance
                if both prefetch windows (and those at the channel and
                connection level) allow it
            connection_global:
                bool: global=false means that the QoS settings should apply
                per-consumer channel; and global=true to mean that the QoS
                settings should apply per-channel.
        """
        request = pamqp.commands.Basic.Qos(
            prefetch_size, prefetch_count, connection_global
        )
        return await self._write_frame_awaiting_response('basic_qos', self.channel_id, request, no_wait=False)

    async def basic_qos_ok(self, frame):
        future = self._get_waiter('basic_qos')
        future.set_result(True)
        logger.debug("Qos ok")

    async def basic_server_nack(self, frame, delivery_tag=None):
        if delivery_tag is None:
            delivery_tag = frame.delivery_tag
        fut = self._get_waiter('basic_server_ack_{}'.format(delivery_tag))
        logger.debug('Received nack for delivery tag %r', delivery_tag)
        fut.set_exception(exceptions.PublishFailed(delivery_tag))

    def new_consumer(
        self,
        queue_name='',
        consumer_tag='',
        no_local=False,
        no_ack=False,
        exclusive=False,
        no_wait=False,
        arguments=None
    ):
        """Starts the consumption of message from a queue.

            Usage::

                async with chan.new_consumer(queue_name="my_queue") \
                        as listener:
                    async for body, envelope, properties in listener:
                        await process_message(body, envelope, properties)

            Arguments:
                queue_name:
                    str, the queue to receive message from
                consumer_tag:
                    str, optional consumer tag
                no_local:
                    bool, if set the server will not send messages to the
                    connection that published them.
                no_ack:
                    bool, if set the server does not expect
                    acknowledgements for messages
                exclusive:
                    bool, request exclusive consumer access, meaning only
                    this consumer can access the queue
                no_wait:
                    bool, if set, the server will not respond to the method
                arguments:
                    dict, AMQP arguments to be passed to the server

        If no callback is given, return an iterable which returns (message,
        envelope, properties) triples.

        Otherwise, the callback function will be called with these three
        arguments, once for each message. Callbacks may be simple functions,
        async coroutines, or AnyIO tasks.

        In either case, unless you have set :param:`no_ack` to ``True``,
        your code is responsible for calling :meth:`basic_client_ack` or
        :meth:`basic_client_nack` on the envelope's :attribute:`delivery_tag`.
        """
        # If a consumer tag was not passed, create one
        consumer_tag = consumer_tag or 'ctag%i.%s' % (self.channel_id, uuid.uuid4().hex)

        if arguments is None:
            arguments = {}

        return BasicListener(
            self,
            queue_name=queue_name,
            consumer_tag=consumer_tag,
            no_local=no_local,
            no_ack=no_ack,
            exclusive=exclusive,
            no_wait=no_wait,
            arguments=arguments
        )

    async def basic_consume(
        self,
        callback,
        queue_name='',
        consumer_tag='',
        no_local=False,
        no_ack=False,
        exclusive=False,
        no_wait=False,
        arguments=None
    ):
        """Starts the consumption of message from a queue.
        The callback will be called each time we're receiving a message.

            Arguments:
                callback:
                    coroutine, the callback to be executed for each message.
                queue_name:
                    str, the queue to receive message from
                consumer_tag:
                    str, optional consumer tag
                no_local:
                    bool, if set the server will not send messages to the
                    connection that published them.
                no_ack:
                    bool, if set the server does not expect
                    acknowledgements for messages
                exclusive:
                    bool, request exclusive consumer access, meaning only
                    this consumer can access the queue
                no_wait:
                    bool, if set, the server will not respond to the method
                arguments:
                    dict, AMQP arguments to be passed to the server

        The callback function will be called with three arguments,
        once for each message. Callbacks may be simple functions, async
        coroutines, or AnyIO tasks.

            Callback Args:
                Message:
                    The body of the AMQP message.
                Envelope:
                    An instance of :class:`async_amqp.envelope.Envelope`.
                Properties:
                    An instance of :class:`async_amqp.properties.Properties`.

        Unless you have set :param:`no_ack` to ``True``, your code is
        responsible for calling :meth:`basic_client_ack` or
        :meth:`basic_client_nack` on the envelope's
        :attribute:`delivery_tag`.
        """
        # If a consumer tag was not passed, create one
        consumer_tag = consumer_tag or 'ctag%i.%s' % (self.channel_id, uuid.uuid4().hex)

        if arguments is None:
            arguments = {}

        request = pamqp.commands.Basic.Consume(
            queue=queue_name,
            consumer_tag=consumer_tag,
            no_local=no_local,
            no_ack=no_ack,
            exclusive=exclusive,
            nowait=no_wait,
            arguments=arguments
        )

        self.consumer_callbacks[consumer_tag] = callback
        self.last_consumer_tag = consumer_tag

        return_value = await self._write_frame_awaiting_response('basic_consume' + consumer_tag, self.channel_id, request, no_wait)
        if no_wait:
            return_value = {'consumer_tag': consumer_tag}
        else:
            self._ctag_events[consumer_tag].set()
        return return_value

    async def basic_consume_ok(self, frame):
        results = {
            'consumer_tag': frame.consumer_tag
        }
        future = self._get_waiter('basic_consume' + frame.consumer_tag)
        future.set_result(results)
        self._ctag_events[frame.consumer_tag] = anyio.Event()

    async def basic_deliver(self, frame):
        consumer_tag = frame.consumer_tag
        delivery_tag = frame.delivery_tag
        is_redeliver = frame.redelivered
        exchange_name = frame.exchange
        routing_key = frame.routing_key
        channel, content_header_frame = await self.protocol.get_frame()

        buffer = io.BytesIO()
        while (buffer.tell() < content_header_frame.body_size):
            _channel, content_body_frame = await self.protocol.get_frame()
            buffer.write(content_body_frame.value)

        body = buffer.getvalue()
        envelope = Envelope(consumer_tag, delivery_tag, exchange_name, routing_key, is_redeliver)
        properties = amqp_properties.from_pamqp(content_header_frame.properties)

        callback = self.consumer_callbacks[consumer_tag]

        event = self._ctag_events.get(consumer_tag)
        if event:
            await event.wait()
            del self._ctag_events[consumer_tag]

        res = callback(self, body, envelope, properties)
        if inspect.iscoroutine(res):
            res = await res

    async def server_basic_cancel(self, frame):
        # https://www.rabbitmq.com/consumer-cancel.html
        consumer_tag = frame.consumer_tag
        _no_wait = frame.nowait
        self.cancelled_consumers.add(consumer_tag)
        logger.info("consume cancelled received")
        callback = self.consumer_callbacks.get(consumer_tag, None)
        if callback is None:
            pass
        elif not inspect.iscoroutinefunction(callback):
            callback(self, None, None, None)
        else:
            await callback(self, None, None, None)


    async def basic_cancel(self, consumer_tag, no_wait=False):
        request = pamqp.commands.Basic.Cancel(consumer_tag, no_wait)
        return await self._write_frame_awaiting_response('basic_cancel', self.channel_id, request, no_wait=no_wait)

    async def basic_cancel_ok(self, frame):
        results = {
            'consumer_tag': frame.consumer_tag,
        }
        future = self._get_waiter('basic_cancel')
        future.set_result(results)
        logger.debug("Cancel ok")

    async def basic_return(self, frame):
        reply_code = frame.reply_code
        reply_text = frame.reply_text
        exchange_name = frame.exchange
        routing_key = frame.routing_key
        channel, content_header_frame = await self.protocol.get_frame()

        buffer = io.BytesIO()
        while buffer.tell() < content_header_frame.body_size:
            _channel, content_body_frame = await self.protocol.get_frame()
            buffer.write(content_body_frame.value)

        body = buffer.getvalue()
        envelope = ReturnEnvelope(reply_code, reply_text,
                                  exchange_name, routing_key)
        properties = amqp_properties.from_pamqp(content_header_frame.properties)
        if self._q_w is None:
            # they have set mandatory bit, but aren't reading
            logger.warning("You don't iterate the channel for returned messages!")
        else:
            await self._q_w.send((body, envelope, properties))

    async def basic_get(self, queue_name='', no_ack=False):
        request = pamqp.commands.Basic.Get(queue=queue_name, no_ack=no_ack)
        return await self._write_frame_awaiting_response('basic_get', self.channel_id, request, no_wait=False)

    async def basic_get_ok(self, frame):
        data = {
            'delivery_tag': frame.delivery_tag,
            'redelivered': frame.redelivered,
            'exchange_name': frame.exchange,
            'routing_key': frame.routing_key,
            'message_count': frame.message_count,
        }
        _channel, content_header_frame = await self.protocol.get_frame()

        buffer = io.BytesIO()
        while (buffer.tell() < content_header_frame.body_size):
            _channel, content_body_frame = await self.protocol.get_frame()
            buffer.write(content_body_frame.value)

        data['message'] = buffer.getvalue()
        data['properties'] = amqp_properties.from_pamqp(content_header_frame.properties)
        future = self._get_waiter('basic_get')
        future.set_result(data)

    async def basic_get_empty(self, frame):
        future = self._get_waiter('basic_get')
        future.set_exception(exceptions.EmptyQueue)

    async def basic_client_ack(self, delivery_tag, multiple=False):
        request = pamqp.commands.Basic.Ack(delivery_tag, multiple)
        async with self._write_lock:
            await self._write_frame(self.channel_id, request)

    async def basic_client_nack(self, delivery_tag, multiple=False, requeue=True):
        request = pamqp.commands.Basic.Nack(delivery_tag, multiple, requeue)
        async with self._write_lock:
            await self._write_frame(self.channel_id, request)

    async def basic_server_ack(self, frame):
        delivery_tag = frame.delivery_tag
        fut = self._get_waiter('basic_server_ack_{}'.format(delivery_tag))
        logger.debug('Received ack for delivery tag %s', delivery_tag)
        fut.set_result(True)

    async def basic_reject(self, delivery_tag, requeue=False):
        request = pamqp.commands.Basic.Reject(delivery_tag, requeue)
        async with self._write_lock:
            await self._write_frame(self.channel_id, request)

    async def basic_recover_async(self, requeue=True):
        request = pamqp.commands.Basic.RecoverAsync(requeue)
        async with self._write_lock:
            await self._write_frame(self.channel_id, request)

    async def basic_recover(self, requeue=True):
        request = pamqp.commands.Basic.Recover(requeue)
        return await self._write_frame_awaiting_response('basic_recover', self.channel_id, request, no_wait=False)

    async def basic_recover_ok(self, frame):
        future = self._get_waiter('basic_recover')
        future.set_result(True)
        logger.debug("Cancel ok")


#
# convenient aliases
#

    queue = queue_declare
    exchange = exchange_declare

    async def publish(
        self,
        payload,
        exchange_name,
        routing_key,
        properties=None,
        mandatory=False,
        immediate=False
    ):
        if properties is None:
            properties = {}
        if not isinstance(payload,(bytes,bytearray)):
            raise TypeError("Payload must be bytes")

        async with self._write_lock:
            if self.publisher_confirms:
                delivery_tag = next(self.delivery_tag_iter)
                fut = self._set_waiter('basic_server_ack_{}'.format(delivery_tag))

            method_request = pamqp.commands.Basic.Publish(
                exchange=exchange_name,
                routing_key=routing_key,
                mandatory=mandatory,
                immediate=immediate
            )
            await self._write_frame(self.channel_id, method_request, drain=False)

            properties = pamqp.commands.Basic.Properties(**properties)
            header_request = pamqp.header.ContentHeader(
                body_size=len(payload), properties=properties
            )

            await self._write_frame(self.channel_id, header_request, drain=False)

            # split the payload

            frame_max = self.protocol.server_frame_max or len(payload)
            for chunk in (payload[0 + i:frame_max + i] for i in range(0, len(payload), frame_max)):
                content_request = pamqp.body.ContentBody(chunk)
                await self._write_frame(self.channel_id, content_request, drain=False)

            await self.protocol._drain()

        if self.publisher_confirms:
            await fut()

    async def confirm_select(self, *, no_wait=False):
        if self.publisher_confirms:
            raise ValueError('publisher confirms already enabled')
        request = pamqp.commands.Confirm.Select(nowait=no_wait)

        return await self._write_frame_awaiting_response('confirm_select', self.channel_id, request, no_wait)

    async def confirm_select_ok(self, frame):
        self.publisher_confirms = True
        self.delivery_tag_iter = count(1)
        fut = self._get_waiter('confirm_select')
        fut.set_result(True)
        logger.debug("Confirm selected")



"""
    Amqp basic tests for recover methods
"""

import pytest

from . import testcase


class TestRecover(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_basic_recover_async(self, channel):
        with pytest.deprecated_call():
            await channel.basic_recover_async(requeue=True)

    @pytest.mark.xfail(msg="server doesn't like that")
    @pytest.mark.trio
    async def test_basic_recover_async_no_requeue(self, channel):
        with pytest.deprecated_call():
            await channel.basic_recover_async(requeue=False)

    @pytest.mark.trio
    async def test_basic_recover(self, channel):
        result = await channel.basic_recover(requeue=True)
        assert result



#!/usr/bin/env python
"""
    Worker example from the 2nd tutorial
"""

import anyio
import async_amqp


async def callback(channel, body, envelope, properties):
    print(" [x] Received %r" % body)
    await anyio.sleep(body.count(b'.'))
    print(" [x] Done")
    await channel.basic_client_ack(delivery_tag=envelope.delivery_tag)


async def worker():
    async with async_amqp.connect_amqp() as protocol:

        channel = await protocol.channel()

        await channel.queue(queue_name='task_queue', durable=True)
        await channel.basic_qos(
            prefetch_count=1, prefetch_size=0, connection_global=False
        )
        await channel.basic_consume(callback, queue_name='task_queue')
        while True:
            await anyio.sleep(99999)


anyio.run(worker)



#!/usr/bin/env python
"""
    Rabbitmq.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-five-python.html
"""

import anyio
import async_amqp

import sys


async def exchange_routing_topic():
    try:
        async with async_amqp.connect_amqp() as protocol:

            channel = await protocol.channel()
            exchange_name = 'topic_logs'
            message = ' '.join(sys.argv[2:]) or 'Hello World!'
            routing_key = sys.argv[1] if len(
                sys.argv
            ) > 1 else 'anonymous.info'

            await channel.exchange(exchange_name, 'topic')

            await channel.publish(
                message, exchange_name=exchange_name, routing_key=routing_key
            )
            print(" [x] Sent %r" % message)

    except async_amqp.AmqpClosedConnection:
        print("closed connections")
        return


anyio.run(exchange_routing_topic)



"""
    Hello world `send.py` example implementation using async_amqp.
    See the documentation for more informations.

"""

import anyio
import async_amqp


async def send():
    async with async_amqp.connect_amqp() as protocol:
        channel = await protocol.channel()

        await channel.queue_declare(queue_name='hello')

        await channel.basic_publish(
            payload='Hello World!', exchange_name='', routing_key='hello'
        )

        print(" [x] Sent 'Hello World!'")


anyio.run(send)



import pytest
import anyio
from async_amqp.protocol import OPEN, CLOSED

from . import testcase


@pytest.mark.skip(reason="epoll ignores closing")
class TestConnectionLost(testcase.RabbitTestCase):

    _multiprocess_can_split_ = True

    @pytest.mark.trio
    async def test_connection_lost(self, amqp):
        async with amqp.new_channel() as channel:
            assert amqp.state == OPEN
            assert channel.is_open
            # os.close(amqp._stream.socket.fileno()) # does not work w/ epoll
            # this should have the same effect as the tcp connection being lost
            await amqp._stream.aclose()

            with anyio.fail_after(1):
                await amqp.connection_closed.wait()
            assert amqp.state == CLOSED
            assert not channel.is_open



import anyio
import pytest

from . import testcase


class TestPublish(testcase.RabbitTestCase):

    _multiprocess_can_split_ = True

    @pytest.mark.trio
    async def test_publish(self, channel):
        # declare
        await channel.queue_declare("q", exclusive=True, no_wait=False)
        await channel.exchange_declare("e", "fanout")
        await channel.queue_bind("q", "e", routing_key='')

        # publish
        await channel.publish(b"coucou", "e", routing_key='')

        await self.check_messages(channel.protocol, "q", 1)

    @pytest.mark.trio
    async def test_empty_publish(self, channel):
        # declare
        await channel.queue_declare("q", exclusive=True, no_wait=False)
        await channel.exchange_declare("e", "fanout")
        await channel.queue_bind("q", "e", routing_key='')

        # publish
        await channel.publish(b"", "e", routing_key='')

        await self.check_messages(channel.protocol, "q", 1)

    @pytest.mark.trio
    async def test_big_publish(self, channel):
        # declare
        await channel.queue_declare("q", exclusive=True, no_wait=False)
        await channel.exchange_declare("e", "fanout")
        await channel.queue_bind("q", "e", routing_key='')

        # publish
        await channel.publish(b"a" * 1000000, "e", routing_key='')

        await self.check_messages(channel.protocol, "q", 1)

    @pytest.mark.trio
    async def test_confirmed_publish(self, channel):
        # declare
        await channel.confirm_select()
        assert channel.publisher_confirms
        await channel.queue_declare("q", exclusive=True, no_wait=False)
        await channel.exchange_declare("e", "fanout")
        await channel.queue_bind("q", "e", routing_key='')

        # publish
        await channel.publish(b"coucou", "e", routing_key='')

        await self.check_messages(channel.protocol, "q", 1)


    @pytest.mark.skip("This callback doesn't exist in async_amqp")
    @pytest.mark.trio
    async def test_return_from_publish(self, channel):
        called = trio.Event()

        async def logger(task_status=trio.TASK_STATUS_IGNORED):
            task_status.started()
            async for a,b,c in channel:
                called.set()

        async def sender(task_status=trio.TASK_STATUS_IGNORED):
            task_status.started()

            # declare
            await channel.exchange_declare("e", "topic")

            # publish
            await channel.publish("coucou", "e", routing_key="not.found",
                                  mandatory=True)
 
        with trio.fail_after(1):
            async with trio.open_nursery() as n:
                await n.start(logger)
                await n.start(sender)
                await called.wait()
                n.cancel_scope.cancel()
 



"""
    Amqp Protocol
"""

import anyio
import errno
import logging
from math import inf
import socket
import ssl
try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager
import pamqp
import pamqp.commands
from anyio.abc import SocketAttribute

from . import channel as amqp_channel
from . import constants as amqp_constants
from . import frame as amqp_frame
from . import exceptions

logger = logging.getLogger(__name__)

CONNECTING, OPEN, CLOSING, CLOSED = range(4)
READ_BUF_SIZE = 1024


class ChannelContext:
    """This class is returned by :meth:`AmqpProtocol:new_channel`.
    It is responsible for creating a new channel when used as a context
    manager.
    """

    def __init__(self, conn, kwargs):
        self.conn = conn
        self.kwargs = kwargs

    async def __aenter__(self):
        self.channel = await self.conn.channel(**self.kwargs)
        return self.channel

    async def __aexit__(self, *tb):
        if not self.channel.is_open:
            return
        with anyio.move_on_after(2,shield=True):
            try:
                await self.channel.close()
            except exceptions.AmqpClosedConnection:
                pass

    def __enter__(self):
        raise RuntimeError("This is an async-only context manager.")

    def __exit__(self, *tb):
        raise RuntimeError("This is an async-only context manager.")


class AmqpProtocol:
    """The AMQP protocol for anyio.
    """

    CHANNEL_FACTORY = amqp_channel.Channel
    CHANNEL_CONTEXT = ChannelContext

    _close_reason = None

    def __init__(
        self,
        nursery,
        host='localhost',
        port=None,
        ssl=False,
        verify_ssl=True,
        login='guest',
        password='guest',
        virtualhost='/',
        channel_max=None,
        frame_max=None,
        heartbeat=None,
        client_properties=None,
        login_method='PLAIN',
        insist=False
    ):
        """Defines our new protocol instance

        Args:
            host:
                the host to connect to
            port:
                broker port
            login:
                login
            password:
                password
            virtualhost:
                AMQP virtualhost to use for this connection
            ssl:
                the SSL context to use
            login_method:
                AMQP auth method
            insist:
                Insist on connecting to a server

            kwargs:
                Arguments to be given to the protocol_factory instance

            channel_max:
                specifies highest channel number that the server permits.
                Usable channel numbers are in the range 1..channel-max.
                Zero indicates no specified limit.
            frame_max:
                the largest frame size that the server proposes for the
                connection, including frame header and end-byte. The client
                can negotiate a lower value.
                Zero means that the server does not impose any specific
                limit but may reject very large frames if it cannot
                allocate resources for them.
            heartbeat:
                the delay, in seconds, of the connection heartbeat that the
                server wants. Zero means the server does not want a
                heartbeat.
            client_properties:
                dict, client-props to tune the client identification
        """

        self._reader_scope = None
        self._writer_scope = None

        self._nursery = nursery
        self.client_properties = client_properties or {}
        self.connection_tunning = {}
        if channel_max is not None:
            self.connection_tunning['channel_max'] = channel_max
        if frame_max is not None:
            self.connection_tunning['frame_max'] = frame_max
        if heartbeat is not None:
            self.connection_tunning['heartbeat'] = heartbeat

        if login_method != 'PLAIN':
            logger.warning('login_method %s is not supported, falling back to PLAIN', login_method)

        self._host = host
        self._port = port
        self._ssl = ssl
        self._virtualhost = virtualhost
        self._login_method = login_method
        self._insist = insist
        self._auth = {
            'LOGIN': login,
            'PASSWORD': password,
        }

    @property
    def nursery(self):
        return self._nursery

    async def ensure_open(self):
        # Raise a suitable exception if the connection isn't open.
        # Handle cases from the most common to the least common.

        if self.state == OPEN:
            return

        if self.state == CLOSED:
            raise exceptions.AmqpClosedConnection()

        # If the closing handshake is in progress, let it complete.
        if self.state == CLOSING:
            await self.wait_closed()
            raise exceptions.AmqpClosedConnection()

        # Control may only reach this point in buggy third-party subclasses.
        assert self.state == CONNECTING
        raise exceptions.AsyncAmqpException("connection isn't established yet.")

    async def _drain(self):
        return

        # with (await self._drain_lock):
        #    # drain() cannot be called concurrently by multiple coroutines:
        #    # http://bugs.python.org/issue29930. Remove this lock when no
        #    # version of Python where this bugs exists is supported anymore.
        #    await self._stream_writer.drain()

    async def _write_frame(self, channel_id, request, drain=True):
        # Doesn't actually write frame, pushes it for _writer_loop task to
        # pick it up.
        data = pamqp.frame.marshal(request, channel_id)
        await self._send_queue_w.send(data)

    async def _writer_loop(self, *, task_status):
        with anyio.CancelScope(shield=True) as scope:
            self._writer_scope = scope
            task_status.started()
            while self.state != CLOSED:
                if self.server_heartbeat:
                    timeout = self.server_heartbeat / 2
                else:
                    timeout = inf

                with anyio.move_on_after(timeout) as timeout_scope:
                    data = await self._send_queue_r.receive()
                if timeout_scope.cancel_called:
                    await self.send_heartbeat()
                    continue

                try:
                    await self._stream.send(data)
                except (anyio.ClosedResourceError, BrokenPipeError):
                    # raise exceptions.AmqpClosedConnection(self) from None
                    # the reader will raise the error also
                    return

    async def close(self, no_wait=False):
        """Close connection (and all channels)"""
        if self.state == CLOSED:
            return
        if self.state == CLOSING:
            if not no_wait:
                await self.wait_closed()
            return

        try:
            self.state = CLOSING
            got_close = self.connection_closed.is_set()
            self.connection_closed.set()
            if not got_close:
                await self._close_channels()

                # If the closing handshake is in progress, let it complete.
                request = pamqp.commands.Connection.Close(
                    reply_code=0,
                    reply_text='',
                    class_id=0,
                    method_id=0
                )
                try:
                    await self._write_frame(0, request)
                except anyio.ClosedResourceError:
                    pass
                except Exception:
                    logger.exception("Error while closing")
                else:
                    if not no_wait and self.server_heartbeat:
                        with anyio.move_on_after(self.server_heartbeat / 2):
                            await self.wait_closed()

        except BaseException as exc:
            with anyio.fail_after(2, shield=True):
                await self._close_channels(exception=exc)
            raise

        finally:
            with anyio.fail_after(2, shield=True):
                try:
                    self._cancel_all()
                    await self._stream.aclose()
                finally:
                    self._nursery = None
                    self.state = CLOSED

    async def wait_closed(self):
        await self.connection_closed.wait()

    async def close_ok(self, frame):
        logger.debug("Recv close ok")
        await self._stream.aclose()

    def __enter__(self):
        raise TypeError("You need to use an async context")

    def __exit__(self, a, b, c):
        raise TypeError("You need to use an async context")

    async def __aenter__(self):
        self.connection_closed = anyio.Event()
        self.state = CONNECTING
        self.version_major = None
        self.version_minor = None
        self.server_properties = None
        self.server_mechanisms = None
        self.server_locales = None
        self.server_heartbeat = None
        self.channels = {}
        self.server_frame_max = None
        self.server_channel_max = None
        self.channels_ids_ceil = 0
        self.channels_ids_free = set()
        self._send_queue_w,self._send_queue_r = anyio.create_memory_object_stream(1)

        if self._ssl:
            if self._ssl is True:
                ssl_context = ssl.create_default_context()
            else:
                ssl_context = self._ssl

        port = self._port
        if port is None:
            if self._ssl:
                port = 5671
            else:
                port = 5672

        if self._ssl:
            stream = await anyio.connect_tcp(self._host, port, ssl_context=ssl_context, autostart_tls=True)
        else:
            stream = await anyio.connect_tcp(self._host, port)

        stream.extra(SocketAttribute.raw_socket).setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        from anyio.streams.buffered import BufferedByteReceiveStream
        self._stream = stream
        self._rstream = BufferedByteReceiveStream(stream)

        # the writer loop needs to run from the beginning
        await self._nursery.start(self._writer_loop)

        try:
            await self._stream.send(amqp_constants.PROTOCOL_HEADER)

            # Wait 'start' method from the server
            await self.dispatch_frame()
            if self.version_major is None:
                raise RuntimeError("Server didn't start with a START packet")

            client_properties = {
                'capabilities': {
                    'consumer_cancel_notify': True,
                    'connection.blocked': False,
                },
            }
            client_properties.update(self.client_properties)

            # waiting reply start with credentions and co
            await self.start_ok(client_properties, 'AMQPLAIN', self._auth, self.server_locales)

            # wait for a "tune" reponse
            await self.dispatch_frame()
            if self.server_channel_max is None:
                raise RuntimeError("Server didn't send a TUNE packet")

            tune_ok = {
                'channel_max':
                    self.connection_tunning.get('channel_max', self.server_channel_max),
                'frame_max':
                    self.connection_tunning.get('frame_max', self.server_frame_max),
                'heartbeat':
                    self.connection_tunning.get('heartbeat', self.server_heartbeat),
            }
            # "tune" the connexion with max channel, max frame, heartbeat
            await self.tune_ok(**tune_ok)

            # update connection tunning values
            self.server_frame_max = tune_ok['frame_max']
            self.server_channel_max = tune_ok['channel_max']
            self.server_heartbeat = tune_ok['heartbeat']

            # open a virtualhost
            await self.open(self._virtualhost, capabilities='', insist=self._insist)

            # wait for open-ok
            await self.dispatch_frame()
            if self.state != OPEN:
                if self._close_reason is not None:
                    raise exceptions.AmqpClosedConnection(self._close_reason['text'])
                else:
                    raise exceptions.AmqpClosedConnection()

            # read the other server's responses asynchronously
            await self._nursery.start(self._reader_loop)

        except BaseException as exc:
            with anyio.fail_after(2, shield=True):
                await self.close(no_wait=True)
            raise

        return self

    async def __aexit__(self, typ, exc, tb):
        with anyio.move_on_after(2, shield=True):
            await self.close()

    async def get_frame(self):
        """Read the frame, and only decode its header

        """
        try:
            channel, frame = await amqp_frame.read(self._rstream)
        except ConnectionResetError:
            raise exceptions.AmqpClosedConnection(self) from None
        except EnvironmentError as err:
            if err.errno == errno.EBADF:
                raise exceptions.AmqpClosedConnection(self) from None
            raise
        except anyio.ClosedResourceError:
            raise exceptions.AmqpClosedConnection(self) from None

        return channel, frame

    async def dispatch_frame(self, frame_channel=None, frame=None):
        """Dispatch the received frame to the corresponding handler"""

        method_dispatch = {
            pamqp.commands.Connection.Close.name: self.server_close,
            pamqp.commands.Connection.CloseOk.name: self.close_ok,
            pamqp.commands.Connection.Tune.name: self.tune,
            pamqp.commands.Connection.Start.name: self.start,
            pamqp.commands.Connection.OpenOk.name: self.open_ok,
        }
        if frame is None:
            frame_channel, frame = await self.get_frame()

        if isinstance(frame, pamqp.heartbeat.Heartbeat):
            return

        if frame_channel:
            channel = self.channels.get(frame_channel)
            if channel is not None:
                await channel.dispatch_frame(frame)
            else:
                logger.info("Unknown channel %s", frame_channel)
            return

        if frame.name not in method_dispatch:
            logger.info("frame %s is not handled", frame.name)
            return
        await method_dispatch[frame.name](frame)

    def release_channel_id(self, channel_id):
        """Called from the channel instance, it relase a previously used
        channel_id
        """
        self.channels_ids_free.add(channel_id)

    @property
    def channels_ids_count(self):
        return self.channels_ids_ceil - len(self.channels_ids_free)

    async def _close_channels(self, reply_code=None, reply_text=None, exception=None):
        """Cleanly close channels

            Args:
                reply_code:     int, the amqp error code
                reply_text:     str, the text associated with the error_code
                exception:      the exception responsible of this error

        """
        if exception is None:
            exception = exceptions.ChannelClosed(reply_code, reply_text)

        for channel in self.channels.values():
            await channel.connection_closed(reply_code, reply_text, exception)

    async def _reader_loop(self, *, task_status):
        with anyio.CancelScope(shield=True) as scope:
            self._reader_scope = scope
            try:
                task_status.started()
                while True:
                    try:
                        if self._stream is None:
                            raise exceptions.AmqpClosedConnection

                        if self.server_heartbeat:
                            timeout = self.server_heartbeat * 2
                        else:
                            timeout = inf

                        with anyio.fail_after(timeout):
                            try:
                                channel, frame = await self.get_frame()
                            except anyio.ClosedResourceError:
                                # the stream is now *really* closed …
                                return
                        try:
                            await self.dispatch_frame(channel, frame)
                        except Exception as exc:
                            # We want to raise this exception so that the
                            # nursery ends the protocol, but we need keep
                            # going for now (need to process the close-OK
                            # message). Thus we start a new task that
                            # raises the actual error, somewhat later.
                            if self._nursery is None:
                                raise

                            async def owch(exc):
                                await anyio.sleep(0.01)
                                raise exc

                            logger.error("Queue %r", exc)
                            self._nursery.start_soon(owch, exc)

                    except TimeoutError:
                        self.connection_closed.set()
                        raise exceptions.HeartbeatTimeoutError(self) from None
                    except exceptions.AmqpClosedConnection as exc:
                        logger.debug("Remote closed connection")
                        if self.state in (CLOSING, CLOSED):
                            return
                        raise
            finally:
                self._reader_scope = None
                self.connection_closed.set()

    async def send_heartbeat(self):
        """Sends an heartbeat message.
        It can be an ack for the server or the client willing to check for the
        connexion timeout
        """
        request = pamqp.heartbeat.Heartbeat()
        await self._write_frame(0, request)

    # Amqp specific methods
    async def start(self, frame):
        """Method sent from the server to begin a new connection"""
        self.version_major = frame.version_major
        self.version_minor = frame.version_minor
        self.server_properties = frame.server_properties
        self.server_mechanisms = frame.mechanisms
        self.server_locales = frame.locales

    async def start_ok(self, client_properties, mechanism, auth, locale):
        class StartOk(pamqp.commands.Connection.StartOk):
            _response = 'table'

        request = StartOk(
            client_properties=client_properties,
            mechanism=mechanism,
            locale=locale,
            response=auth,
        )
        await self._write_frame(0, request)

    async def server_close(self, frame):
        """The server is closing the connection"""
        self.state = CLOSING
        reply_code = frame.reply_code
        reply_text = frame.reply_text
        class_id = frame.class_id
        method_id = frame.method_id
        self._close_reason = dict(text=reply_text, code=reply_code, class_id=class_id, method_id=method_id)
        logger.warning(
            "Server closed connection: %s, code=%s, class_id=%s, method_id=%s", reply_text,
            reply_code, class_id, method_id
        )
        await self._close_channels(reply_code, reply_text)
        await self._close_ok()

    async def _close_ok(self):
        request = pamqp.commands.Connection.CloseOk()
        await self._write_frame(0, request)
        await anyio.sleep(0)  # give the write task one shot to send the frame
        if self._nursery is not None:
            self._cancel_all()

    def _cancel_all(self):
        if self._reader_scope is not None:
            self._reader_scope.cancel()
        if self._writer_scope is not None:
            self._writer_scope.cancel()
        if self._nursery is not None:
            self._nursery.cancel_scope.cancel()

    async def tune(self, frame):
        self.server_channel_max = frame.channel_max
        self.server_frame_max = frame.frame_max
        self.server_heartbeat = frame.heartbeat

    async def tune_ok(self, channel_max, frame_max, heartbeat):
        request = pamqp.commands.Connection.TuneOk(
            channel_max, frame_max, heartbeat
        )
        await self._write_frame(0, request)

    async def secure_ok(self, login_response):
        pass

    async def open(self, virtual_host, capabilities='', insist=False):
        """Open connection to virtual host."""
        request = pamqp.commands.Connection.Open(
            virtual_host, capabilities, insist
        )
        await self._write_frame(0, request)

    async def open_ok(self, frame):
        self.state = OPEN
        logger.debug("Recv open ok")

    #
    # async_amqp public methods
    #

    def new_channel(self, **kwargs):
        return self.CHANNEL_CONTEXT(self, kwargs)

    async def channel(self, **kwargs):
        """Factory to create a new channel

        """
        await self.ensure_open()
        try:
            channel_id = self.channels_ids_free.pop()
        except KeyError as ex:
            assert self.server_channel_max is not None, \
                'connection channel-max tuning not performed'
            # channel-max = 0 means no limit
            if self.server_channel_max and \
                    self.channels_ids_ceil > self.server_channel_max:
                raise exceptions.NoChannelAvailable() from ex
            self.channels_ids_ceil += 1
            channel_id = self.channels_ids_ceil
        channel = self.CHANNEL_FACTORY(self, channel_id, **kwargs)
        self.channels[channel_id] = channel
        await channel.open()
        return channel


@asynccontextmanager
async def connect_amqp(*args, protocol=AmqpProtocol, **kwargs):
    async with anyio.create_task_group() as nursery:
        amqp = protocol(nursery, *args, **kwargs)
        try:
            async with amqp:
                yield amqp
        except anyio.BrokenResourceError as ex:
            raise exceptions.AmqpClosedConnection from ex
        finally:
            with anyio.fail_after(2, shield=True):
                amqp._cancel_all()




"""AsyncAmqp tests utilities

Provides the test case to simplify testing
"""

import anyio
import inspect
import logging
import os
import time
import uuid
import pytest
from functools import wraps
try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager

import pyrabbit2 as pyrabbit

from . import testcase
from async_amqp import exceptions, connect_amqp
from async_amqp.channel import Channel
from async_amqp.protocol import AmqpProtocol, OPEN
from anyio._core._compat import DeprecatedAwaitable

logger = logging.getLogger(__name__)


def use_full_name(f, arg_names):
    sig = inspect.signature(f)
    for arg_name in arg_names:
        if arg_name not in sig.parameters:
            raise ValueError(
                '%s is not a valid argument name for function %s' % (arg_name, f.__qualname__)
            )

    @wraps(f)
    def wrapper(self, *args, **kw):
        ba = sig.bind_partial(self, *args, **kw)
        for param in sig.parameters.values():
            if param.name in arg_names and param.name in ba.arguments:
                ba.arguments[param.name] = self.full_name(ba.arguments[param.name])
        return f(*(ba.args), **(ba.kwargs))

    return wrapper


class ProxyChannel(Channel):
    exchange_declare = use_full_name(Channel.exchange_declare, ['exchange_name'])
    exchange_delete = use_full_name(Channel.exchange_delete, ['exchange_name'])
    queue_declare = use_full_name(Channel.queue_declare, ['queue_name'])
    queue_delete = use_full_name(Channel.queue_delete, ['queue_name'])
    queue_bind = use_full_name(Channel.queue_bind, ['queue_name', 'exchange_name'])
    queue_unbind = use_full_name(Channel.queue_unbind, ['queue_name', 'exchange_name'])
    queue_purge = use_full_name(Channel.queue_purge, ['queue_name'])

    exchange_bind = use_full_name(
        Channel.exchange_bind, ['exchange_source', 'exchange_destination']
    )
    exchange_unbind = use_full_name(
        Channel.exchange_unbind, ['exchange_source', 'exchange_destination']
    )
    publish = use_full_name(Channel.publish, ['exchange_name'])
    basic_get = use_full_name(Channel.basic_get, ['queue_name'])
    basic_consume = use_full_name(Channel.basic_consume, ['queue_name'])

    def full_name(self, name):
        conn = self.protocol
        if conn is None:
            return name
        return conn.full_name(name)


_seq = 0


class ProxyAmqpProtocol(AmqpProtocol):
    test_seq = None
    CHANNEL_FACTORY = ProxyChannel

    def __init__(self, *args, test_seq=None, **kwargs):
        if test_seq is None:
            global _seq
            _seq += 1
            test_seq = _seq
        super().__init__(*args, **kwargs)
        self.test_seq = test_seq

    @property
    def shortname(self):
        return "tc_%03d_" % self.test_seq

    async def _stop(self):
        pass

    def full_name(self, name):
        if self.is_full_name(name):
            return name
        return self.shortname + name

    def local_name(self, name):
        if self.is_full_name(name):
            return name[len(self.shortname):]
        return name

    def is_full_name(self, name):
        return name.startswith(self.shortname)


def reset_vhost():
    host = os.environ.get('AMQP_HOST', 'localhost')
    port = int(os.environ.get('AMQP_PORT', 5672))
    vhost = os.environ.get('AMQP_VHOST', 'test' + str(uuid.uuid4()))
    http_client = pyrabbit.api.Client(
        '%s:%s' % (host, 10000 + port), 'guest', 'guest', timeout=20
    )

    try:
        http_client.delete_vhost(vhost)
    except Exception:  # pylint: disable=broad-except
        pass

    http_client.create_vhost(vhost)
    http_client.set_vhost_permissions(
        vname=vhost, username='guest', config='.*', rd='.*', wr='.*',
    )


def connect(*a, **kw):
    return connect_amqp(*a, protocol=ProxyAmqpProtocol, **kw)


class FakeScope:
    def __init__(self, scope):
        self.scope = scope

    def cancel(self):
        self.scope.cancel()
        return DeprecatedAwaitable(self.cancel)


class TaskGroup:
    def __init__(self, nursery) -> None:
        self._nursery = nursery
        self.cancel_scope = FakeScope(nursery.cancel_scope)
             
    def start_soon(self, func, *args, name=None) -> None:
        self._nursery.start_soon(func, *args, name=name)

    async def start(self, func, *args, name=None) -> None:
        return await self._nursery.start(func, *args, name=name)



@pytest.fixture
async def amqp(request, nursery):
    reset_vhost()
    async with ProxyAmqpProtocol(
        nursery=TaskGroup(nursery),
        host=os.environ.get('AMQP_HOST', 'localhost'),
        port=int(os.environ.get('AMQP_PORT', 5672)),
        virtualhost=os.environ.get('AMQP_VHOST', 'test' + str(uuid.uuid4())),
    ) as conn:
        yield conn


@pytest.fixture
async def channel(request, nursery):
    reset_vhost()
    # XXX using another async fixture does not work yet
    async with ProxyAmqpProtocol(
        nursery=TaskGroup(nursery),
        host=os.environ.get('AMQP_HOST', 'localhost'),
        port=int(os.environ.get('AMQP_PORT', 5672)),
        virtualhost=os.environ.get('AMQP_VHOST', 'test' + str(uuid.uuid4())),
    ) as conn:
        async with conn.new_channel() as channel:
            yield channel


class RabbitTestCase:
    """TestCase with a rabbit running in background"""

    RABBIT_TIMEOUT = 1.0
    amqp = None

    def setup(self):
        self.host = os.environ.get('AMQP_HOST', 'localhost')
        self.port = int(os.environ.get('AMQP_PORT', 5672))
        self.vhost = os.environ.get('AMQP_VHOST', 'test' + str(uuid.uuid4()))
        self.http_client = pyrabbit.api.Client(
            '%s:%s' % (self.host, 10000 + self.port), 'guest', 'guest', timeout=20
        )

        self.amqps = []
        self.exchanges = {}
        self.queues = {}
        self.transports = []

    async def start(self, rc=None):
        pass

    async def stop(self):
        for queue_name, channel in self.queues.values():
            logger.debug('Delete queue %s', self.full_name(queue_name))
            await self.safe_queue_delete(queue_name, channel)
        for exchange_name, channel in self.exchanges.values():
            logger.debug('Delete exchange %s', self.full_name(exchange_name))
            await self.safe_exchange_delete(exchange_name, channel)
        for amqp in self.amqps:
            if amqp.state != OPEN:
                continue
            logger.debug('Kill off amqp %s', amqp)
            await amqp._nursery.cancel_scope.cancel()
            await amqp.close()

    def teardown(self):
        pass
#       try:
#           self.http_client.delete_vhost(self.vhost)
#       except Exception as exc:  # pylint: disable=broad-except
#           pass

    def reset_vhost(self):
        reset_vhost()  # global

    def server_version(self, amqp):
        server_version = tuple(int(x) for x in amqp.server_properties['version'].split('.'))
        return server_version

    async def check_exchange_exists(self, exchange_name):
        """Check if the exchange exist"""
        try:
            await self.exchange_declare(exchange_name, passive=True)
        except exceptions.ChannelClosed:
            return False

        return True

    async def assertExchangeExists(self, exchange_name):
        if not self.check_exchange_exists(exchange_name):
            self.fail("Exchange {} does not exists".format(exchange_name))

    async def check_queue_exists(self, queue_name):
        """Check if the queue exist"""
        try:
            await self.queue_declare(queue_name, passive=True)
        except exceptions.ChannelClosed:
            return False

        return True

    async def assertQueueExists(self, queue_name):
        if not self.check_queue_exists(queue_name):
            self.fail("Queue {} does not exists".format(queue_name))

    async def list_queues(self, amqp, vhost=None, fully_qualified_name=False, delay=None):
        # wait for the http client to get the correct state of the queue
        if delay is None:
            delay = int(os.environ.get('AMQP_REFRESH_TIME', 1.1))
        await anyio.sleep(delay)
        queues_list = self.http_client.get_queues(vhost=vhost or self.vhost)
        queues = {}
        for queue_info in queues_list:
            queue_name = queue_info['name']
            if fully_qualified_name is False:
                queue_name = amqp.local_name(queue_info['name'])
                queue_info['name'] = queue_name

            queues[queue_name] = queue_info
        return queues

    async def check_messages(self, amqp, queue_name, num_msg):
        for x in range(20):
            try:
                queues = await self.list_queues(amqp)
                assert queue_name in queues
                q = queues[queue_name]
                try:
                    q['messages_ready_ram'] == 1
                except (KeyError, AssertionError):
                    try:
                        assert q['messages'] == 1
                    except (KeyError, AssertionError):
                        assert q['message_stats']['publish'] == 1
            except (KeyError, AssertionError) as exc:
                ex = exc
            else:
                break
            await anyio.sleep(0.5)
        else:
            raise ex

    async def safe_queue_delete(self, queue_name, channel):
        """Delete the queue but does not raise any exception if it fails

        The operation has a timeout as well.
        """
        full_queue_name = channel.protocol.full_name(queue_name)
        try:
            await channel.queue_delete(full_queue_name, no_wait=False)
        except TimeoutError:
            logger.warning('Timeout on queue %s deletion', full_queue_name, exc_info=True)
        except Exception:  # pylint: disable=broad-except
            logger.exception('Unexpected error on queue %s deletion', full_queue_name)

    async def safe_exchange_delete(self, exchange_name, channel=None):
        """Delete the exchange but does not raise any exception if it fails

        The operation has a timeout as well.
        """
        channel = channel or self.channel
        full_exchange_name = self.full_name(exchange_name)
        try:
            await channel.exchange_delete(full_exchange_name, no_wait=False)
        except TimeoutError:
            logger.warning('Timeout on exchange %s deletion', full_exchange_name, exc_info=True)
        except Exception:  # pylint: disable=broad-except
            logger.exception('Unexpected error on exchange %s deletion', full_exchange_name)

    async def queue_declare(self, queue_name, *args, channel=None, safe_delete_before=True, **kw):
        channel = channel or self.channel
        if safe_delete_before:
            await self.safe_queue_delete(queue_name, channel=channel)
        # prefix queue_name with the test name
        full_queue_name = self.full_name(queue_name)
        try:
            rep = await channel.queue_declare(full_queue_name, *args, **kw)

        finally:
            self.queues[queue_name] = (queue_name, channel)
        return rep

    async def exchange_declare(
        self, exchange_name, *args, channel=None, safe_delete_before=True, **kw
    ):
        channel = channel or self.channel
        if safe_delete_before:
            await self.safe_exchange_delete(exchange_name, channel=channel)
        # prefix exchange name
        full_exchange_name = self.full_name(exchange_name)
        try:
            rep = await channel.exchange_declare(full_exchange_name, *args, **kw)
        finally:
            self.exchanges[exchange_name] = (exchange_name, channel)
        return rep

    @asynccontextmanager
    async def create_amqp(self, vhost=None, test_seq=None):
        async with testcase.connect(
            host=self.host, port=self.port, virtualhost=vhost or self.vhost, test_seq=test_seq
        ) as protocol:
            self.amqps.append(protocol)
            yield protocol



"""
    Tests the heartbeat methods
"""

import anyio
import pytest

from async_amqp import exceptions

from . import testcase


class TestHeartbeat(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_recv_heartbeat(self):
        conn = testcase.connect(virtualhost=self.vhost,)
        self.reset_vhost()
        with pytest.raises(exceptions.HeartbeatTimeoutError):
            async with conn as amqp:

                async def mock_send():
                    self.send_called += 1

                self.send_called = 0

                amqp.send_heartbeat = mock_send
                amqp.server_heartbeat = 0.01

                async with amqp.new_channel():
                    # this ensures that the send and recv loops use the new
                    # heartbeat value
                    await anyio.sleep(0.1)
                    assert False, "not reached"
        assert self.send_called > 2



#!/usr/bin/env python
"""
    Rabbitmq.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-five-python.html

"""

import anyio
import async_amqp

import sys


async def callback(channel, body, envelope, properties):
    print(
        "consumer {} received {} ({})".format(
            envelope.consumer_tag, body, envelope.delivery_tag
        )
    )


async def receive_log():
    try:
        async with async_amqp.connect_amqp() as protocol:

            channel = await protocol.channel()
            exchange_name = 'topic_logs'

            await channel.exchange(exchange_name, 'topic')

            result = await channel.queue(
                queue_name='', durable=False, auto_delete=True
            )
            queue_name = result['queue']

            binding_keys = sys.argv[1:]
            if not binding_keys:
                print("Usage: %s [binding_key]..." % (sys.argv[0],))
                sys.exit(1)

            for binding_key in binding_keys:
                await channel.queue_bind(
                    exchange_name='topic_logs',
                    queue_name=queue_name,
                    routing_key=binding_key
                )

            print(' [*] Waiting for logs. To exit press CTRL+C')

            await channel.basic_consume(callback, queue_name=queue_name)

            while True:
                await anyio.sleep(99999)
    except async_amqp.AmqpClosedConnection:
        print("closed connections")
        return


anyio.run(receive_log)



"""
    RPC server, async_amqp implementation of RPC examples from RabbitMQ tutorial
"""

import anyio
import async_amqp


def fib(n):
    if n == 0:
        return 0
    elif n == 1:
        return 1
    else:
        return fib(n - 1) + fib(n - 2)


async def on_request(channel, body, envelope, properties):
    n = int(body)

    print(" [.] fib(%s)" % n)
    response = fib(n)

    await channel.basic_publish(
        payload=str(response),
        exchange_name='',
        routing_key=properties.reply_to,
        properties={
            'correlation_id': properties.correlation_id,
        },
    )

    await channel.basic_client_ack(delivery_tag=envelope.delivery_tag)


async def rpc_server():
    async with async_amqp.connect_amqp() as protocol:

        channel = await protocol.channel()

        await channel.queue_declare(queue_name='rpc_queue')
        await channel.basic_qos(
            prefetch_count=1, prefetch_size=0, connection_global=False
        )
        await channel.basic_consume(on_request, queue_name='rpc_queue')
        print(" [x] Awaiting RPC requests")
        while True:
            await anyio.sleep(99999)


anyio.run(rpc_server)



"""
    Amqp Future implementation
"""

import anyio
import logging
import outcome
from anyio._core._compat import DeprecatedAwaitable

logger = logging.getLogger(__name__)


class FutureCancelled(Exception):
    pass


class Future:
    def __init__(self, channel, rpc_name):
        self.channel = channel
        self.rpc_name = rpc_name
        self.event = anyio.Event()
        self.result = None
        channel._add_future(self)

    async def __call__(self):
        await self.event.wait()
        return self.result.unwrap()

    def set_result(self, value):
        if self.event.is_set():
            raise RuntimeError("future already set")
        self.result = outcome.Value(value)
        self.event.set()
        return DeprecatedAwaitable(self.set_result)

    def set_exception(self, exc):
        if self.event.is_set():
            raise RuntimeError("future already set")
        if isinstance(exc, type):
            exc = exc()
        self.result = outcome.Error(exc)
        self.event.set()
        return DeprecatedAwaitable(self.set_exception)

    def cancel(self):
        try:
            raise FutureCancelled()
        except FutureCancelled as exc:
            self.set_exception(exc)
        return DeprecatedAwaitable(self.cancel)

    def done(self):
        return self.event.is_set()



from urllib.parse import urlparse
try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager

from .exceptions import *  # pylint: disable=wildcard-import  # noqa: F401,F403
from .protocol import AmqpProtocol  # noqa: F401

from . import protocol
connect_amqp = protocol.connect_amqp


@asynccontextmanager
async def connect_from_url(url, **kwargs):
    """Connect to the AMQP using a single url parameter.

        @url:    amqp:// or amqps:// URL with connection parameters
        @kwargs: Further arguments for async_amqp.connect_amqp()

        Usage:
            async with connect_from_url(
                "amqp://user:password@hostname:port/vhost"
            ) as amqp:
                await do_whatever(amqp)
    """
    url = urlparse(url)

    if url.scheme not in ('amqp', 'amqps'):
        raise ValueError('Invalid protocol %s, valid protocols are amqp or amqps' % url.scheme)

    if url.hostname:
        kwargs['host'] = url.hostname
    if url.port:
        kwargs['port'] = url.port
    if url.username:
        kwargs['login'] = url.username
    if url.password:
        kwargs['password'] = url.password
    async with connect_amqp(
        virtualhost=(url.path[1:] if len(url.path) > 1 else '/'),
        ssl=(url.scheme == 'amqps'),
        **kwargs
    ) as amqp:
        yield amqp



# pylint: disable=redefined-builtin
from .constants import MESSAGE_PROPERTIES


class Properties:
    """Class for basic message properties"""
    __slots__ = tuple(MESSAGE_PROPERTIES)

    def __init__(
        self,
        content_type=None,
        content_encoding=None,
        headers=None,
        delivery_mode=None,
        priority=None,
        correlation_id=None,
        reply_to=None,
        expiration=None,
        message_id=None,
        timestamp=None,
        message_type=None,
        user_id=None,
        app_id=None,
        cluster_id=None
    ):
        self.content_type = content_type
        self.content_encoding = content_encoding
        self.headers = headers
        self.delivery_mode = delivery_mode
        self.priority = priority
        self.correlation_id = correlation_id
        self.reply_to = reply_to
        self.expiration = expiration
        self.message_id = message_id
        self.timestamp = timestamp
        self.message_type = message_type
        self.user_id = user_id
        self.app_id = app_id
        self.cluster_id = cluster_id


def from_pamqp(instance):
    props = Properties()
    props.content_type = instance.content_type
    props.content_encoding = instance.content_encoding
    props.headers = instance.headers
    props.delivery_mode = instance.delivery_mode
    props.priority = instance.priority
    props.correlation_id = instance.correlation_id
    props.reply_to = instance.reply_to
    props.expiration = instance.expiration
    props.message_id = instance.message_id
    props.timestamp = instance.timestamp
    props.message_type = instance.message_type
    props.user_id = instance.user_id
    props.app_id = instance.app_id
    props.cluster_id = instance.cluster_id
    return props



"""
    Hello world `receive.py` example implementation using async_amqp.
    See the documentation for more informations.
"""

import anyio
import async_amqp


async def callback(channel, body, envelope, properties):
    print(" [x] Received %r" % body)


async def receive():
    try:
        async with async_amqp.connect_amqp() as protocol:

            channel = await protocol.channel()

            await channel.queue_declare(queue_name='hello')

            await channel.basic_consume(callback, queue_name='hello')

            while True:
                await anyio.sleep(99999)
    except async_amqp.AmqpClosedConnection:
        print("closed connections")
        return


anyio.run(receive)



"""
    Defines our constants
"""

PROTOCOL_DEFAULT_TIMEOUT = 60
PROTOCOL_DEFAULT_PORT = 5672
PROTOCOL_HEADER = b'AMQP\x01\x01\x00\x09'

MAX_CHANNELS = 65535

# protocol
TYPE_METHOD = 1
TYPE_HEADER = 2
TYPE_BODY = 3
TYPE_HEARTBEAT = 8

FRAME_END = b'\xce'

# classes
CLASS_CONNECTION = 10
CLASS_CHANNEL = 20
CLASS_EXCHANGE = 40
CLASS_QUEUE = 50
CLASS_BASIC = 60
CLASS_TX = 90
CLASS_CONFIRM = 85

CONNECTION_START = 10
CONNECTION_START_OK = 11
CONNECTION_SECURE = 20
CONNECTION_SECURE_OK = 21
CONNECTION_TUNE = 30
CONNECTION_TUNE_OK = 31
CONNECTION_OPEN = 40
CONNECTION_OPEN_OK = 41
CONNECTION_CLOSE = 50
CONNECTION_CLOSE_OK = 51

CHANNEL_OPEN = 10
CHANNEL_OPEN_OK = 11
CHANNEL_FLOW = 20
CHANNEL_FLOW_OK = 21
CHANNEL_CLOSE = 40
CHANNEL_CLOSE_OK = 41

EXCHANGE_DECLARE = 10
EXCHANGE_DECLARE_OK = 11
EXCHANGE_DELETE = 20
EXCHANGE_DELETE_OK = 21
EXCHANGE_BIND = 30
EXCHANGE_BIND_OK = 31
EXCHANGE_UNBIND = 40
EXCHANGE_UNBIND_OK = 51

QUEUE_DECLARE = 10
QUEUE_DECLARE_OK = 11
QUEUE_BIND = 20
QUEUE_BIND_OK = 21
QUEUE_UNBIND = 50
QUEUE_UNBIND_OK = 51
QUEUE_PURGE = 30
QUEUE_PURGE_OK = 31
QUEUE_DELETE = 40
QUEUE_DELETE_OK = 41

BASIC_QOS = 10
BASIC_QOS_OK = 11
BASIC_CONSUME = 20
BASIC_CONSUME_OK = 21
BASIC_CANCEL = 30
BASIC_CANCEL_OK = 31
BASIC_PUBLISH = 40
BASIC_RETURN = 50
BASIC_DELIVER = 60
BASIC_GET = 70
BASIC_GET_OK = 71
BASIC_GET_EMPTY = 72
BASIC_ACK = 80
BASIC_REJECT = 90
BASIC_RECOVER_ASYNC = 100
BASIC_RECOVER = 110
BASIC_RECOVER_OK = 111
BASIC_NACK = 120

TX_SELECT = 10
TX_SELECT_OK = 11
TX_COMMIT = 20
TX_COMMIT_OK = 21
TX_ROLLBACK = 30
TX_ROLLBACK_OK = 31

CONFIRM_SELECT = 10
CONFIRM_SELECT_OK = 11

MESSAGE_PROPERTIES = (
    'content_type', 'content_encoding', 'headers', 'delivery_mode', 'priority', 'correlation_id',
    'reply_to', 'expiration', 'message_id', 'timestamp', 'message_type', 'user_id', 'app_id', 'cluster_id',
)

FLAG_CONTENT_TYPE = (1 << 15)
FLAG_CONTENT_ENCODING = (1 << 14)
FLAG_HEADERS = (1 << 13)
FLAG_DELIVERY_MODE = (1 << 12)
FLAG_PRIORITY = (1 << 11)
FLAG_CORRELATION_ID = (1 << 10)
FLAG_REPLY_TO = (1 << 9)
FLAG_EXPIRATION = (1 << 8)
FLAG_MESSAGE_ID = (1 << 7)
FLAG_TIMESTAMP = (1 << 6)
FLAG_TYPE = (1 << 5)
FLAG_USER_ID = (1 << 4)
FLAG_APP_ID = (1 << 3)
FLAG_CLUSTER_ID = (1 << 2)



#!/usr/bin/env python

import anyio
import async_amqp

import sys


async def new_task():
    try:
        async with async_amqp.connect_amqp() as protocol:

            channel = await protocol.channel()

            await channel.queue('task_queue', durable=True)

            message = ' '.join(sys.argv[1:]) or "Hello World!"

            await channel.basic_publish(
                payload=message,
                exchange_name='',
                routing_key='task_queue',
                properties={
                    'delivery_mode': 2,
                },
            )
            print(" [x] Sent %r" % message,)

    except async_amqp.AmqpClosedConnection:
        print("closed connections")
        return


anyio.run(new_task)



"""
    async_amqp exceptions
"""


class AsyncAmqpException(Exception):
    pass


class HeartbeatTimeoutError(AsyncAmqpException):
    """No heartbeat has been received"""
    pass


class ConfigurationError(AsyncAmqpException):
    """unused, kept for compatibility"""
    pass


class AmqpClosedConnection(AsyncAmqpException):
    pass


class SynchronizationError(AsyncAmqpException):
    pass


class EmptyQueue(AsyncAmqpException):
    pass


class NoChannelAvailable(AsyncAmqpException):
    """There is no room left for more channels"""
    pass


class ChannelClosed(AsyncAmqpException):
    def __init__(self, code=0, message='Channel is closed'):
        super().__init__(code, message)
        self.code = code
        self.message = message


class DuplicateConsumerTag(AsyncAmqpException):
    def __repr__(self):
        return ('The consumer tag specified already exists for this ' 'channel: %s' % self.args[0])


class ConsumerCancelled(AsyncAmqpException):
    def __repr__(self):
        return ('The consumer %s has been cancelled' % self.args[0])


class PublishFailed(AsyncAmqpException):
    def __init__(self, delivery_tag):
        super().__init__(delivery_tag)
        self.delivery_tag = delivery_tag

    def __repr__(self):
        return 'Publish failed because a nack was ' \
            'received for delivery_tag {}'.format(  # noqa: E122
            self.delivery_tag
        )



"""
    Amqp basic class tests
"""

import anyio
import struct
import pytest

from . import testcase
from async_amqp import exceptions
from async_amqp import properties


class TestQos(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_basic_qos_default_args(self, channel):
        result = await channel.basic_qos()
        assert result

    @pytest.mark.trio
    async def test_basic_qos(self, channel):
        result = await channel.basic_qos(
            prefetch_size=0, prefetch_count=100, connection_global=False
        )

        assert result

    @pytest.mark.trio
    async def test_basic_qos_prefetch_size(self):
        self.reset_vhost()
        conn = testcase.connect(virtualhost=self.vhost,)
        async with conn as amqp:
            async with amqp.new_channel() as channel:
                with pytest.raises(exceptions.ChannelClosed) as cm:
                    await channel.basic_qos(
                        prefetch_size=10, prefetch_count=100, connection_global=False
                    )

        assert cm.value.code == 540

    @pytest.mark.trio
    async def test_basic_qos_wrong_values(self):
        self.reset_vhost()
        conn = testcase.connect(virtualhost=self.vhost,)
        async with conn as amqp:
            async with amqp.new_channel() as channel:
                with pytest.raises(TypeError):
                    await channel.basic_qos(
                        prefetch_size=100000, prefetch_count=1000000000, connection_global=False
                    )


class TestBasicCancel(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_basic_cancel(self, channel):
        async def callback(channel, body, envelope, _properties):
            pass

        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        await channel.queue_declare(queue_name)
        await channel.exchange_declare(exchange_name, type_name='direct')
        await channel.queue_bind(queue_name, exchange_name, routing_key='')
        result = await channel.basic_consume(callback, queue_name=queue_name)
        result = await channel.basic_cancel(result['consumer_tag'])

        result = await channel.publish(b"payload", exchange_name, routing_key='')

        await anyio.sleep(1)
        result = await channel.queue_declare(queue_name, passive=True)
        assert result['message_count'] == 1
        assert result['consumer_count'] == 0

    @pytest.mark.trio
    async def test_basic_cancel_unknown_ctag(self, channel):
        result = await channel.basic_cancel("unknown_ctag")
        assert result


class TestBasicGet(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_basic_get(self, channel):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''

        await channel.queue_declare(queue_name)
        await channel.exchange_declare(exchange_name, type_name='direct')
        await channel.queue_bind(queue_name, exchange_name, routing_key=routing_key)

        await channel.publish(b"payload", exchange_name, routing_key=routing_key)

        result = await channel.basic_get(queue_name)
        assert result['routing_key'] == routing_key
        # XXX # assert not result['redelivered']
        assert 'delivery_tag' in result
        assert result['exchange_name'] == channel.protocol.full_name(exchange_name)
        assert result['message'] == b'payload'
        assert isinstance(result['properties'], properties.Properties)

    @pytest.mark.trio
    async def test_basic_get_empty(self, channel):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''
        await channel.queue_declare(queue_name)
        await channel.exchange_declare(exchange_name, type_name='direct')
        await channel.queue_bind(queue_name, exchange_name, routing_key=routing_key)

        with pytest.raises(exceptions.EmptyQueue):
            await channel.basic_get(queue_name)


class TestBasicDelivery(testcase.RabbitTestCase):
    async def publish(self, amqp, queue_name, exchange_name, routing_key, payload):
        async with amqp.new_channel() as channel:
            await channel.queue_declare(queue_name, exclusive=False, no_wait=False)
            await channel.exchange_declare(exchange_name, type_name='fanout')
            await channel.queue_bind(queue_name, exchange_name, routing_key=routing_key)
            await channel.publish(payload, exchange_name, queue_name)

    @pytest.mark.trio
    async def test_ack_message(self, amqp):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''

        await self.publish(amqp, queue_name, exchange_name, routing_key, b"payload")

        qfuture = anyio.Event()

        async def qcallback(channel, body, envelope, _properties):
            qfuture.set()
            self.test_result = envelope

        async with amqp.new_channel() as channel:
            await channel.basic_consume(qcallback, queue_name=queue_name)
            await qfuture.wait()
            envelope = self.test_result

            await channel.basic_client_ack(envelope.delivery_tag)

    @pytest.mark.trio
    async def test_basic_nack(self, amqp):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''

        await self.publish(amqp, queue_name, exchange_name, routing_key, b"payload")

        qfuture = anyio.Event()

        async with amqp.new_channel() as channel:

            async def qcallback(channel, body, envelope, _properties):
                await channel.basic_client_nack(
                    envelope.delivery_tag, multiple=True, requeue=False
                )
                qfuture.set()

            await channel.basic_consume(qcallback, queue_name=queue_name)
            await qfuture.wait()

    @pytest.mark.trio
    async def test_basic_nack_norequeue(self, amqp):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''

        await self.publish(amqp, queue_name, exchange_name, routing_key, b"payload")

        qfuture = anyio.Event()

        async with amqp.new_channel() as channel:

            async def qcallback(channel, body, envelope, _properties):
                await channel.basic_client_nack(envelope.delivery_tag, requeue=False)
                qfuture.set()

            await channel.basic_consume(qcallback, queue_name=queue_name)
            await qfuture.wait()

    @pytest.mark.trio
    async def test_basic_nack_requeue(self, amqp):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''

        await self.publish(amqp, queue_name, exchange_name, routing_key, b"payload")

        qfuture = anyio.Event()
        called = False

        async with amqp.new_channel() as channel:

            async def qcallback(channel, body, envelope, _properties):
                nonlocal called
                if not called:
                    called = True
                    await channel.basic_client_nack(envelope.delivery_tag, requeue=True)
                else:
                    await channel.basic_client_ack(envelope.delivery_tag)
                    qfuture.set()

            await channel.basic_consume(qcallback, queue_name=queue_name)
            await qfuture.wait()

    @pytest.mark.trio
    async def test_basic_reject(self, amqp):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''
        await self.publish(amqp, queue_name, exchange_name, routing_key, b"payload")

        qfuture = anyio.Event()

        async def qcallback(channel, body, envelope, _properties):
            qfuture.set()
            self.test_result = envelope

        async with amqp.new_channel() as channel:
            await channel.basic_consume(qcallback, queue_name=queue_name)
            await qfuture.wait()
            envelope = self.test_result

            await channel.basic_reject(envelope.delivery_tag)



"""
    Server received requests handling tests

"""

import pytest

from . import testcase


class TestServerBasicCancel(testcase.RabbitTestCase):
    _multiprocess_can_split_ = True

    @pytest.mark.trio
    async def test_cancel_whilst_consuming(self, channel):
        queue_name = 'queue_name'
        await channel.queue_declare(queue_name)

        # None is non-callable.  We want to make sure the callback is
        # unregistered and never called.
        await channel.basic_consume(None)
        await channel.queue_delete(queue_name)



import anyio
import pytest

from . import testcase
from async_amqp import exceptions

from async_amqp.properties import Properties


class TestConsume(testcase.RabbitTestCase):

    _multiprocess_can_split_ = True

#   def setup(self):
#       super().setup()
#       self.consume_future = anyio.Event()

    async def callback(self, channel, body, envelope, properties):
        self.consume_result = (body, envelope, properties)
        self.consume_future.set()

    async def get_callback_result(self):
        await self.consume_future.wait()
        result = self.consume_result
        self.consume_future = anyio.Event()
        return result

    @pytest.mark.trio
    async def test_wrong_callback_argument(self):
        self.consume_future = anyio.Event()
        def badcallback():
            pass

        self.reset_vhost()
        proto = testcase.connect(virtualhost=self.vhost,)
        with pytest.raises(TypeError):
            async with proto as amqp:
                async with amqp.new_channel() as chan:
                    await chan.queue_declare("q", exclusive=True, no_wait=False)
                    await chan.exchange_declare("e", "fanout")
                    await chan.queue_bind("q", "e", routing_key='')

                    # get a different channel
                    async with amqp.new_channel() as channel:
                        # publish
                        await channel.publish(
                            "coucou",
                            "e",
                            routing_key='',
                        )

                        # assert there is a message to consume
                        await self.check_messages(amqp, "q", 1)

                        # start consume
                        await channel.basic_consume(badcallback, queue_name="q")
                        await anyio.sleep(1)

    @pytest.mark.trio
    async def test_consume(self, amqp):
        self.consume_future = anyio.Event()
        # declare
        async with amqp.new_channel() as channel:
            await channel.queue_declare("q", exclusive=True, no_wait=False)
            await channel.exchange_declare("e", "fanout")
            await channel.queue_bind("q", "e", routing_key='')

            # get a different channel
            async with amqp.new_channel() as channel:
                # publish
                await channel.publish(
                    b"coucou",
                    "e",
                    routing_key='',
                )

                # start consume
                await channel.basic_consume(self.callback, queue_name="q")

                # get one
                body, envelope, properties = await self.get_callback_result()
                assert envelope.consumer_tag is not None
                assert envelope.delivery_tag is not None
                assert b"coucou" == body
                assert isinstance(properties, Properties)

    @pytest.mark.trio
    async def test_big_consume(self, amqp):
        self.consume_future = anyio.Event()
        # declare
        async with amqp.new_channel() as channel:
            await channel.queue_declare("q", exclusive=True, no_wait=False)
            await channel.exchange_declare("e", "fanout")
            await channel.queue_bind("q", "e", routing_key='')

            # get a different channel
            async with amqp.new_channel() as channel:
                # publish
                await channel.publish(
                    b"a" * 1000000,
                    "e",
                    routing_key='',
                )

                # start consume
                await channel.basic_consume(self.callback, queue_name="q")

                # get one
                body, envelope, properties = await self.get_callback_result()
                assert envelope.consumer_tag is not None
                assert envelope.delivery_tag is not None
                assert b"a" * 1000000 == body
                assert isinstance(properties, Properties)

    @pytest.mark.trio
    async def test_consume_multiple_queues(self, amqp):
        self.consume_future = anyio.Event()
        async with amqp.new_channel() as channel:
            await channel.queue_declare("q1", exclusive=True, no_wait=False)
            await channel.queue_declare("q2", exclusive=True, no_wait=False)
            await channel.exchange_declare("e", "direct")
            await channel.queue_bind("q1", "e", routing_key="q1")
            await channel.queue_bind("q2", "e", routing_key="q2")

            # get a different channel
            async with amqp.new_channel() as channel:

                q1_future = anyio.Event()

                async def q1_callback(channel, body, envelope, properties):
                    self.q1_result = (body, envelope, properties)
                    q1_future.set()

                q2_future = anyio.Event()

                async def q2_callback(channel, body, envelope, properties):
                    self.q2_result = (body, envelope, properties)
                    q2_future.set()

                # start consumers
                result = await channel.basic_consume(q1_callback, queue_name="q1")
                ctag_q1 = result['consumer_tag']
                result = await channel.basic_consume(q2_callback, queue_name="q2")
                ctag_q2 = result['consumer_tag']

                # put message in q1
                await channel.publish(b"coucou1", "e", "q1")

                # get it
                await q1_future.wait()
                body1, envelope1, properties1 = self.q1_result
                assert ctag_q1 == envelope1.consumer_tag
                assert envelope1.delivery_tag is not None
                assert b"coucou1" == body1
                assert isinstance(properties1, Properties)

                # put message in q2
                await channel.publish(b"coucou2", "e", "q2")

                # get it
                await q2_future.wait()
                body2, envelope2, properties2 = self.q2_result
                assert ctag_q2 == envelope2.consumer_tag
                assert b"coucou2" == body2
                assert isinstance(properties2, Properties)

    @pytest.mark.trio
    @pytest.mark.xfail(msg="Needs debugging")
    async def test_duplicate_consumer_tag(self, channel):
        self.consume_future = anyio.Event()
        await channel.queue_declare("q1", exclusive=True, no_wait=False)
        await channel.queue_declare("q2", exclusive=True, no_wait=False)
        await channel.basic_consume(self.callback, queue_name="q1", consumer_tag='tag')

        with pytest.raises(exceptions.ChannelClosed) as cm:
            await channel.basic_consume(self.callback, queue_name="q2", consumer_tag='tag')

        assert cm.value.code == 530

    @pytest.mark.trio
    async def test_consume_callaback_synced(self, amqp):
        self.consume_future = anyio.Event()
        # declare
        async with amqp.new_channel() as channel:
            await channel.queue_declare("q", exclusive=True, no_wait=False)
            await channel.exchange_declare("e", "fanout")
            await channel.queue_bind("q", "e", routing_key='')

            # get a different channel
            async with amqp.new_channel() as channel:

                # publish
                await channel.publish(
                    b"coucou",
                    "e",
                    routing_key='',
                )

                sync_future = anyio.Event()

                async def callback(channel, body, envelope, properties):
                    assert sync_future.is_set()

                await channel.basic_consume(callback, queue_name="q")
                sync_future.set()



import setuptools

description = 'AMQP implementation using anyio'

setuptools.setup(
    name="async_amqp",
    use_scm_version={"version_scheme": "guess-next-dev", "local_scheme": "dirty-tag"},
    author="Matthias Urlichs",
    author_email='matthias@urlichs.de',
    url='https://github.com/python-trio/async_amqp',
    description=description,
    long_description=open('README.rst').read(),
    # download_url='https://pypi.python.org/pypi/async_amqp',
    setup_requires=[
        'pyrabbit',
    ],
    install_requires=[
        'anyio>=3,<4',
        'pamqp>=3,<4',
    ],
    keywords=['asyncio', 'amqp', 'rabbitmq', 'aio', 'trio'],
    packages=[
        'async_amqp',
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Framework :: Trio",
    ],
    platforms='all',
    license='BSD'
)



"""
    Hello world `send.py` example implementation using async_amqp.
    See the documentation for more informations.

    If there is no queue listening for the routing key, the message will
    get returned.

"""

import sys
import anyio
import async_amqp

async def handle_return(channel, body, envelope, properties):
    print('Got a returned message with routing key: {}.\n'
          'Return code: {}\n'
          'Return message: {}\n'
          'exchange: {}'.format(envelope.routing_key, envelope.reply_code,
                                envelope.reply_text, envelope.exchange_name))


async def get_returns(chan):
    # DO NOT await() between these statements
    async for body, envelope, properties in chan:
        await handle_return(channel, body, envelope, properties)


async def send():
    async with async_amqp.connect_amqp() as protocol:
        channel = await protocol.channel()
        await protocol.taskgroup.spawn(get_returns, channel)

        await channel.queue_declare(queue_name='hello')

        await channel.basic_publish(
            payload='Hello World!',
            exchange_name='',
            routing_key='helo',  # typo on purpose, will cause the return
            mandatory=True,
        )

        print(" [x] Sent 'Hello World!'")


anyio.run(send)



#!/usr/bin/env python
"""
    RabbitMQ.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-three-python.html

"""

import anyio
import async_amqp

import sys


async def exchange_routing():
    try:
        async with async_amqp.connect_amqp() as protocol:

            channel = await protocol.channel()
            exchange_name = 'logs'
            message = ' '.join(sys.argv[1:]) or "info: Hello World!"

            await channel.exchange_declare(
                exchange_name=exchange_name, type_name='fanout'
            )
            await channel.basic_publish(
                message, exchange_name=exchange_name, routing_key=''
            )
            print(" [x] Sent %r" % (message,))

    except async_amqp.AmqpClosedConnection:
        print("closed connections")
        return


anyio.run(exchange_routing)



"""
    Amqp exchange class tests
"""

import pytest

from . import testcase
from async_amqp import exceptions


class TestExchangeDeclare(testcase.RabbitTestCase):

    _multiprocess_can_split_ = True

    @pytest.mark.trio
    async def test_exchange_direct_declare(self, channel):
        result = await channel.exchange_declare('exchange_name', type_name='direct')
        assert result

    @pytest.mark.trio
    async def test_exchange_fanout_declare(self, channel):
        result = await channel.exchange_declare('exchange_name', type_name='fanout')
        assert result

    @pytest.mark.trio
    async def test_exchange_topic_declare(self, channel):
        result = await channel.exchange_declare('exchange_name', type_name='topic')
        assert result

    @pytest.mark.trio
    async def test_exchange_headers_declare(self, channel):
        result = await channel.exchange_declare('exchange_name', type_name='headers')
        assert result

    @pytest.mark.trio
    async def test_exchange_declare_wrong_types(self, channel):
        result = await channel.exchange_declare(
            'exchange_name', type_name='headers', auto_delete=True, durable=True
        )
        assert result

        with pytest.raises(exceptions.ChannelClosed):
            result = await channel.exchange_declare(
                'exchange_name', type_name='fanout', auto_delete=False, durable=False
            )

    @pytest.mark.trio
    async def test_exchange_declare_passive(self, channel):
        result = await channel.exchange_declare(
            'exchange_name', type_name='headers', auto_delete=True, durable=True
        )
        assert result
        result = await channel.exchange_declare(
            'exchange_name', type_name='headers', auto_delete=True, durable=True, passive=True
        )
        assert result

        result = await channel.exchange_declare(
            'exchange_name', type_name='headers', auto_delete=False, durable=False, passive=True
        )
        assert result

    @pytest.mark.trio
    async def test_exchange_declare_passive_does_not_exists(self, channel):
        with pytest.raises(exceptions.ChannelClosed) as cm:
            await channel.exchange_declare(
                'non_existant_exchange',
                type_name='headers',
                auto_delete=False,
                durable=False,
                passive=True
            )
        assert cm.value.code == 404

    @pytest.mark.trio
    async def test_exchange_declare_unknown_type(self, channel):
        with pytest.raises(exceptions.ChannelClosed):
            await channel.exchange_declare(
                'non_existant_exchange',
                type_name='unknown_type',
                auto_delete=False,
                durable=False,
                passive=True
            )


class TestExchangeDelete(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_delete(self, channel):
        exchange_name = 'exchange_name'
        await channel.exchange_declare(exchange_name, type_name='direct')
        result = await channel.exchange_delete(exchange_name)
        assert result
        with pytest.raises(exceptions.ChannelClosed) as cm:
            await channel.exchange_declare(exchange_name, type_name='direct', passive=True)

        assert cm.value.code == 404

    @pytest.mark.trio
    async def test_double_delete(self, channel):
        exchange_name = 'exchange_name'
        await channel.exchange_declare(exchange_name, type_name='direct')
        result = await channel.exchange_delete(exchange_name)
        assert result
        if self.server_version(channel.protocol) < (3, 3, 5):
            with pytest.raises(exceptions.ChannelClosed) as cm:
                await channel.exchange_delete(exchange_name)

            assert cm.value.code == 404

        else:
            # weird result from rabbitmq 3.3.5
            result = await channel.exchange_delete(exchange_name)
            assert result


class TestExchangeBind(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_exchange_bind(self, channel):
        await channel.exchange_declare('exchange_destination', type_name='direct')
        await channel.exchange_declare('exchange_source', type_name='direct')

        result = await channel.exchange_bind(
            'exchange_destination', 'exchange_source', routing_key=''
        )

        assert result

    @pytest.mark.trio
    async def test_inexistant_exchange_bind(self, channel):
        with pytest.raises(exceptions.ChannelClosed) as cm:
            await channel.exchange_bind('exchange_destination', 'exchange_source', routing_key='')

        assert cm.value.code == 404


class TestExchangeUnbind(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_exchange_unbind(self, channel):
        ex_source = 'exchange_source'
        ex_destination = 'exchange_destination'
        await channel.exchange_declare(ex_destination, type_name='direct')
        await channel.exchange_declare(ex_source, type_name='direct')

        await channel.exchange_bind(ex_destination, ex_source, routing_key='')

        await channel.exchange_unbind(ex_destination, ex_source, routing_key='')

    @pytest.mark.trio
    async def test_exchange_unbind_reversed(self, channel):
        ex_source = 'exchange_source'
        ex_destination = 'exchange_destination'
        await channel.exchange_declare(ex_destination, type_name='direct')
        await channel.exchange_declare(ex_source, type_name='direct')

        await channel.exchange_bind(ex_destination, ex_source, routing_key='')

        if self.server_version(channel.protocol) < (3, 3, 5):
            with pytest.raises(exceptions.ChannelClosed) as cm:
                result = await channel.exchange_unbind(ex_source, ex_destination, routing_key='')

            assert cm.value.code == 404

        else:
            # weird result from rabbitmq 3.3.5
            result = await channel.exchange_unbind(ex_source, ex_destination, routing_key='')
            assert result



import anyio
import pytest

from . import testcase
from async_amqp import exceptions


class TestClose(testcase.RabbitTestCase):
    def setUp(self):
        super().setUp()
        self.consume_future = anyio.Event()

    @pytest.mark.trio
    async def callback(self, body, envelope, properties):
        self.consume_result = (body, envelope, properties)

    @pytest.mark.trio
    async def get_callback_result(self):
        await self.consume_future.wait()
        self.consume_future = anyio.Event()
        return self.consume_result

    @pytest.mark.trio
    async def test_close(self, channel):
        assert channel.is_open
        await channel.close()
        assert not channel.is_open

    @pytest.mark.trio
    async def test_multiple_close(self, channel):
        await channel.close()
        assert not channel.is_open
        with pytest.raises(exceptions.ChannelClosed):
            await channel.close()

    @pytest.mark.trio
    async def test_cannot_publish_after_close(self, channel):
        await channel.close()
        with pytest.raises(exceptions.ChannelClosed):
            await channel.publish(b"coucou", "my_e", "")

    @pytest.mark.trio
    async def test_cannot_declare_queue_after_close(self, channel):
        await channel.close()
        with pytest.raises(exceptions.ChannelClosed):
            await channel.queue_declare("qq")

    @pytest.mark.trio
    async def test_cannot_consume_after_close(self, channel):
        await channel.queue_declare("q")
        await channel.close()
        with pytest.raises(exceptions.ChannelClosed):
            await channel.basic_consume(self.callback)



#!/usr/bin/env python
"""
    Rabbitmq.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-three-python.html

"""

import anyio
import async_amqp


async def callback(channel, body, envelope, properties):
    print(" [x] %r" % body)


async def receive_log():
    try:
        async with async_amqp.connect_amqp() as protocol:

            channel = await protocol.channel()
            exchange_name = 'logs'

            await channel.exchange(
                exchange_name=exchange_name, type_name='fanout'
            )

            # let RabbitMQ generate a random queue name
            result = await channel.queue(queue_name='', exclusive=True)

            queue_name = result['queue']
            await channel.queue_bind(
                exchange_name=exchange_name,
                queue_name=queue_name,
                routing_key=''
            )

            print(' [*] Waiting for logs. To exit press CTRL+C')

            await channel.basic_consume(
                callback, queue_name=queue_name, no_ack=True
            )

            while True:
                await anyio.sleep(99999)

    except async_amqp.AmqpClosedConnection:
        print("closed connections")
        return


anyio.run(receive_log)



#!/usr/bin/env python
"""
    Rabbitmq.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-four-python.html
"""

import anyio
import async_amqp

import sys


async def exchange_routing():
    try:
        async with async_amqp.connect_amqp() as protocol:

            channel = await protocol.channel()
            exchange_name = 'direct_logs'
            severity = sys.argv[1] if len(sys.argv) > 1 else 'info'
            message = ' '.join(sys.argv[2:]) or 'Hello World!'

            await channel.exchange(exchange_name, 'direct')

            await channel.publish(
                message, exchange_name=exchange_name, routing_key=severity
            )
            print(" [x] Sent %r" % (message,))

    except async_amqp.AmqpClosedConnection:
        print("closed connections")
        return

anyio.run(exchange_routing)



"""
    Test our Protocol class
"""

import pytest
import mock

from . import testcase
from async_amqp import exceptions
from async_amqp import connect_from_url as amqp_from_url
from async_amqp.protocol import OPEN


class TestProtocol(testcase.RabbitTestCase):
    @pytest.mark.trio
    async def test_connect(self):
        self.reset_vhost()
        amqp = testcase.connect(host=self.host, port=self.port, virtualhost=self.vhost)
        async with amqp as protocol:
            assert protocol.state == OPEN

    @pytest.mark.trio
    async def test_connect_products_info(self):
        self.reset_vhost()
        client_properties = {
            'program': 'async_amqp-tests',
            'program_version': '0.1.1',
        }
        amqp = testcase.connect(
            host=self.host, port=self.port, virtualhost=self.vhost,
            client_properties=client_properties,
        )
        async with amqp as protocol:
            assert protocol.client_properties == client_properties

    @pytest.mark.trio
    async def test_connection_unexistant_vhost(self):
        self.reset_vhost()
        with pytest.raises(exceptions.AmqpClosedConnection):
            amqp = testcase.connect(host=self.host, port=self.port, virtualhost='/unexistant')
            async with amqp:
                pass

    @pytest.mark.trio
    async def test_connection_wrong_login_password(self):
        self.reset_vhost()
        with pytest.raises(exceptions.AmqpClosedConnection):
            amqp = testcase.connect(login='wrong', password='wrong')
            async with amqp:
                pass

    @pytest.mark.trio
    async def test_connection_from_url(self):
        self.reset_vhost()
        with mock.patch('async_amqp.connect_amqp') as connect:

            class func:
                def __init__(self):
                    pass

                async def __aenter__(self):
                    return self

                async def __aexit__(self, *tb):
                    pass

            connect.return_value = func()
            res = amqp_from_url('amqp://tom:pass@example.com:7777/myvhost')
            async with res:
                pass
            connect.assert_called_once_with(
                password='pass',
                ssl=False,
                login='tom',
                host='example.com',
                virtualhost='myvhost',
                port=7777,
            )

    @pytest.mark.trio
    async def test_from_url_raises_on_wrong_scheme(self):
        self.reset_vhost()
        with pytest.raises(ValueError):
            async with amqp_from_url('invalid://'):
                assert False, "does not go here"
